# coding: utf-8

import clinica.engine as ce


__author__ = "Arnaud Marcoux"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Arnaud Marcoux", "Michael Bacci"]
__license__ = "See LICENSE.txt file"
__version__ = "1.0.0"
__maintainer__ = "Arnaud Marcoux"
__email__ = "arnaud.marcoux@inria.fr"
__status__ = "Development"


class PetSurfaceCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline.
        """
        self._name = 'pet-surface'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Surface-based processing of PET images:\nhttp://clinica.run/doc/Pipelines/PET_Surface/'

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory. (Filled with results from t1-freesurfer pipeline')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-pt", "--pet_tracer", type=str, default='fdg',
                              help='PET tracer type. Can be fdg or av45 (default: --pet_tracer fdg)')
        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-tsv", "--subjects_sessions_tsv",
                                 help='TSV file containing a list of subjects with their sessions.')
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results')
        clinica_opt.add_argument("-np", "--n_procs",
                                 metavar=('N'), type=int,
                                 help='Number of cores to use when running the pipeline in parallel (default: --n_procs 1).')

    def run_command(self, args):
        from tempfile import mkdtemp
        from clinica.utils.stream import cprint
        from clinica.pipelines.pet_surface.pet_surface_pipeline import PetSurface

        if args.working_directory is None:
            args.working_directory = mkdtemp()

        pipeline = PetSurface(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv))
        pipeline.parameters = {
            'pet_type': args.pet_tracer,
            'wd': self.absolute_path(args.working_directory),
            'n_procs': args.n_procs
        }
        pipeline.base_dir = self.absolute_path(args.working_directory)
        if args.n_procs:
            pipeline.run(plugin='MultiProc', plugin_args={'n_procs': args.n_procs})
        else:
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8

import clinica.pipelines.engine as cpe

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class T1VolumeDartel2MNI(cpe.Pipeline):
    """T1VolumeDartel2MNI - Dartel template to MNI.

    Args:
        input_dir: A BIDS directory.
        output_dir: An empty output directory where CAPS structured data will
        be written.  subjects_sessions_list: The Subjects-Sessions list file
        (in .tsv format).

    Returns:
        A clinica pipeline object containing the T1VolumeDartel2MNI pipeline.

    Raises:


    """

    def __init__(self, bids_directory=None, caps_directory=None, tsv_file=None, name=None, group_id='default'):
        import os
        super(T1VolumeDartel2MNI, self).__init__(bids_directory, caps_directory, tsv_file, name)

        if not group_id.isalnum():
            raise ValueError('Not valid group_id value. It must be composed only by letters and/or numbers')

        self._group_id = group_id
        # Check that group already exists
        if not os.path.exists(os.path.join(os.path.abspath(caps_directory), 'groups', 'group-' + group_id)):
            error_message = group_id \
                            + ' does not exists, please choose an other one (or maybe you need to run t1-spm-dartel).' \
                            + '\nGroups that already exists in your CAPS directory are : \n'
            list_groups = os.listdir(os.path.join(os.path.abspath(caps_directory), 'groups'))
            is_empty = True
            for e in list_groups:
                if e.startswith('group-'):
                    error_message += e + ' \n'
                    is_empty = False
            if is_empty is True:
                error_message += 'NO GROUP FOUND'
            raise ValueError(error_message)

        # Default parameters
        self._parameters = {'tissues': [1, 2, 3],
                            'bounding_box': None,
                            'voxel_size': None,
                            'modulation': True,
                            'fwhm': [8],
                            'atlas_list': ['AAL2', 'LPBA40', 'Neuromorphometrics', 'AICHA', 'Hammers']
                            }

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['apply_to_files', 'flowfield_files', 'template_file']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return ['normalized_files', 'smoothed_normalized_files', 'atlas_statistics']

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """

        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio

        tissue_names = {1: 'graymatter',
                        2: 'whitematter',
                        3: 'csf',
                        4: 'bone',
                        5: 'softtissue',
                        6: 'background'
                        }

        # Segmented Tissues DataGrabber
        # ==============================
        tissues_caps_reader = npe.MapNode(nio.DataGrabber(infields=['subject_id', 'session',
                                                                    'subject_repeat', 'session_repeat',
                                                                    'tissue'],
                                                          outfields=['out_files']),
                                          name="tissues_caps_reader",
                                          iterfield=['subject_id', 'session',
                                                     'subject_repeat', 'session_repeat'])
        tissues_caps_reader.inputs.base_directory = self.caps_directory
        tissues_caps_reader.inputs.template = 'subjects/%s/%s/t1/spm/segmentation/native_space/%s_%s_T1w_segm-%s_probability.nii*'
        tissues_caps_reader.inputs.subject_id = self.subjects
        tissues_caps_reader.inputs.session = self.sessions
        tissues_caps_reader.inputs.tissue = [tissue_names[t] for t in self.parameters['tissues']]
        tissues_caps_reader.inputs.subject_repeat = self.subjects
        tissues_caps_reader.inputs.session_repeat = self.sessions
        tissues_caps_reader.inputs.sort_filelist = False

        # Flow Fields DataGrabber
        # ==============================
        flowfields_caps_reader = npe.Node(nio.DataGrabber(infields=['subject_id', 'session',
                                                                    'subject_repeat', 'session_repeat'],
                                                          outfields=['out_files']),
                                          name="flowfields_caps_reader")
        flowfields_caps_reader.inputs.base_directory = self.caps_directory
        flowfields_caps_reader.inputs.template = 'subjects/%s/%s/t1/spm/dartel/group-' + self._group_id \
                                                 + '/%s_%s_T1w_target-' + self._group_id \
                                                 + '_transformation-forward_deformation.nii*'
        flowfields_caps_reader.inputs.subject_id = self.subjects
        flowfields_caps_reader.inputs.session = self.sessions
        flowfields_caps_reader.inputs.subject_repeat = self.subjects
        flowfields_caps_reader.inputs.session_repeat = self.sessions
        flowfields_caps_reader.inputs.sort_filelist = False

        # Dartel Template DataGrabber
        # ===========================
        template_caps_reader = npe.Node(nio.DataGrabber(infields=['group_id', 'group_id_repeat'],
                                                        outfields=['out_files']),
                                        name="template_caps_reader")
        template_caps_reader.inputs.base_directory = self.caps_directory
        template_caps_reader.inputs.template = 'groups/group-%s/t1/group-%s_template.nii*'
        template_caps_reader.inputs.group_id = self._group_id
        template_caps_reader.inputs.group_id_repeat = self._group_id
        template_caps_reader.inputs.sort_filelist = False

        self.connect([
            (tissues_caps_reader, self.input_node, [('out_files', 'apply_to_files')]),
            (flowfields_caps_reader, self.input_node, [('out_files', 'flowfield_files')]),
            (template_caps_reader, self.input_node, [('out_files', 'template_file')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """
        import os.path as op
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        import re
        from clinica.utils.io import zip_nii

        # Writing normalized images (and smoothed) into CAPS
        # ==================================================
        write_normalized_node = npe.MapNode(name='write_normalized_node',
                                            iterfield=['container', 'normalized_files', 'smoothed_normalized_files'],
                                            interface=nio.DataSink(infields=['normalized_files',
                                                                             'smoothed_normalized_files']))
        write_normalized_node.inputs.base_directory = self.caps_directory
        write_normalized_node.inputs.parameterization = False
        write_normalized_node.inputs.container = ['subjects/' + self.subjects[i] + '/' + self.sessions[i] +
                                                  '/t1/spm/dartel/group-' + self._group_id
                                                  for i in range(len(self.subjects))]
        write_normalized_node.inputs.regexp_substitutions = [
            (r'(.*)c1(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-graymatter_probability\3'),
            (r'(.*)c2(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-whitematter_probability\3'),
            (r'(.*)c3(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-csf_probability\3'),
            (r'(.*)c4(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-bone_probability\3'),
            (r'(.*)c5(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-softtissue_probability\3'),
            (r'(.*)c6(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-background_probability\3'),
            (r'(.*)mw(sub-.*)_probability(\.nii(\.gz)?)$', r'\1\2_space-Ixi549Space_modulated-on_probability\3'),
            (r'(.*)w(sub-.*)_probability(\.nii(\.gz)?)$', r'\1\2_space-Ixi549Space_modulated-off_probability\3'),
            (r'(.*)/normalized_files/(sub-.*)$', r'\1/\2'),
            (r'(.*)/smoothed_normalized_files/(fwhm-[0-9]+mm)_(sub-.*)_probability(\.nii(\.gz)?)$',
             r'\1/\3_\2_probability\4'),
            (r'trait_added', r'')
        ]

        # Writing atlas statistics into CAPS
        # ==================================
        write_atlas_node = npe.MapNode(name='write_atlas_node',
                                            iterfield=['container', 'atlas_statistics'],
                                            interface=nio.DataSink(infields=['atlas_statistics']))
        write_atlas_node.inputs.base_directory = self.caps_directory
        write_atlas_node.inputs.parameterization = False
        write_atlas_node.inputs.container = ['subjects/' + self.subjects[i] + '/' + self.sessions[i] +
                                             '/t1/spm/dartel/group-' + self._group_id + '/atlas_statistics'
                                             for i in range(len(self.subjects))]
        write_atlas_node.inputs.regexp_substitutions = [
            (r'(.*atlas_statistics)/atlas_statistics/mwc1(sub-.*)(_space-.*_map-graymatter_statistics\.tsv)$',
             r'\1/\2\3'),
            (r'(.*atlas_statistics)/atlas_statistics/(m?w)?(sub-.*_T1w).*(_space-.*_map-graymatter_statistics).*(\.tsv)$',
             r'\1/\3\4\5'),
            (r'trait_added', r'')
        ]

        self.connect([
            (self.output_node, write_normalized_node, [(('normalized_files', zip_nii, True), 'normalized_files'),
                                                       (('smoothed_normalized_files', zip_nii, True),
                                                        'smoothed_normalized_files')]),
            (self.output_node, write_atlas_node, [('atlas_statistics', 'atlas_statistics')])
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """

        import os
        import nipype.interfaces.spm as spm
        import nipype.interfaces.matlab as mlab
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.io import unzip_nii
        from clinica.pipelines.t1_volume_dartel2mni.t1_volume_dartel2mni_utils import prepare_flowfields, join_smoothed_files, atlas_statistics, select_gm_images

        spm_home = os.getenv("SPM_HOME")
        mlab_home = os.getenv("MATLABCMD")
        mlab.MatlabCommand.set_default_matlab_cmd(mlab_home)
        mlab.MatlabCommand.set_default_paths(spm_home)

        if 'SPMSTANDALONE_HOME' in os.environ:
            if 'MCR_HOME' in os.environ:
                matlab_cmd = os.path.join(os.environ['SPMSTANDALONE_HOME'],
                                          'run_spm12.sh') \
                             + ' ' + os.environ['MCR_HOME'] \
                             + ' script'
                spm.SPMCommand.set_mlab_paths(matlab_cmd=matlab_cmd, use_mcr=True)
                version = spm.SPMCommand().version
            else:
                raise EnvironmentError('MCR_HOME variable not in environnement. Althought, '
                                       + 'SPMSTANDALONE_HOME has been found')
        else:
            version = spm.Info.getinfo()

        if version:
            if isinstance(version, dict):
                spm_path = version['path']
                if version['name'] == 'SPM8':
                    print('You are using SPM version 8. The recommended version to use with Clinica is SPM 12. '
                          + 'Please upgrade your SPM toolbox.')
                    tissue_map = os.path.join(spm_path, 'toolbox/Seg/TPM.nii')
                elif version['name'] == 'SPM12':
                    tissue_map = os.path.join(spm_path, 'tpm/TPM.nii')
                else:
                    raise RuntimeError('SPM version 8 or 12 could not be found. Please upgrade your SPM toolbox.')
            if isinstance(version, str):
                if float(version) >= 12.7169:
                    tissue_map = os.path.join(str(spm_home), 'spm12_mcr/spm/spm12/tpm/TPM.nii')
                else:
                    raise RuntimeError('SPM standalone version not supported. Please upgrade SPM standalone.')
        else:
            raise RuntimeError('SPM could not be found. Please verify your SPM_HOME environment variable.')

        # Unzipping
        # =========
        unzip_tissues_node = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=unzip_nii),
                                         name='unzip_tissues_node',
                                         iterfield=['in_file'])
        unzip_flowfields_node = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                           output_names=['out_file'],
                                                           function=unzip_nii),
                                            name='unzip_flowfields_node',
                                            iterfield=['in_file'])
        unzip_template_node = npe.Node(nutil.Function(input_names=['in_file'],
                                                      output_names=['out_file'],
                                                      function=unzip_nii),
                                       name='unzip_template_node')

        # DARTEL2MNI Registration
        # =======================
        dartel2mni_node = npe.MapNode(spm.DARTELNorm2MNI(),
                                      name='dartel2MNI',
                                      iterfield=['apply_to_files', 'flowfield_files'])

        if self.parameters['bounding_box'] is not None:
            dartel2mni_node.inputs.bounding_box = self.parameters['bounding_box']
        if self.parameters['voxel_size'] is not None:
            dartel2mni_node.inputs.voxel_size = self.parameters['voxel_size']
        dartel2mni_node.inputs.modulate = self.parameters['modulation']
        dartel2mni_node.inputs.fwhm = 0

        # Smoothing
        # =========
        if self.parameters['fwhm'] is not None and len(self.parameters['fwhm']) > 0:
            smoothing_node = npe.MapNode(spm.Smooth(),
                                         name='smoothing_node',
                                         iterfield=['in_files'])

            smoothing_node.iterables = [('fwhm', [[x, x, x] for x in self.parameters['fwhm']]),
                                        ('out_prefix', ['fwhm-' + str(x) + 'mm_' for x in self.parameters['fwhm']])]
            smoothing_node.synchronize = True

            join_smoothing_node = npe.JoinNode(interface=nutil.Function(input_names=['smoothed_normalized_files'],
                                                                        output_names=['smoothed_normalized_files'],
                                                                        function=join_smoothed_files),
                                               joinsource='smoothing_node',
                                               joinfield='smoothed_normalized_files',
                                               name='join_smoothing_node')
            self.connect([
                (dartel2mni_node, smoothing_node, [('normalized_files', 'in_files')]),
                (smoothing_node, join_smoothing_node, [('smoothed_files', 'smoothed_normalized_files')]),
                (join_smoothing_node, self.output_node, [('smoothed_normalized_files', 'smoothed_normalized_files')])
            ])
        else:
            self.output_node.inputs.smoothed_normalized_files = []

        # Atlas Statistics
        # ================
        atlas_stats_node = npe.MapNode(nutil.Function(input_names=['in_image',
                                                                   'in_atlas_list'],
                                                      output_names=['atlas_statistics'],
                                                      function=atlas_statistics),
                                       name='atlas_stats_node',
                                       iterfield=['in_image'])
        atlas_stats_node.inputs.in_atlas_list = self.parameters['atlas_list']

        # Connection
        # ==========
        self.connect([
            (self.input_node, unzip_tissues_node, [('apply_to_files', 'in_file')]),
            (self.input_node, unzip_flowfields_node, [('flowfield_files', 'in_file')]),
            (self.input_node, unzip_template_node, [('template_file', 'in_file')]),
            (unzip_tissues_node, dartel2mni_node, [('out_file', 'apply_to_files')]),
            (unzip_flowfields_node, dartel2mni_node, [(('out_file', prepare_flowfields, self.parameters['tissues']),
                                                       'flowfield_files')]),
            (unzip_template_node, dartel2mni_node, [('out_file', 'template_file')]),
            (dartel2mni_node, self.output_node, [('normalized_files', 'normalized_files')]),
            (dartel2mni_node, atlas_stats_node, [(('normalized_files', select_gm_images), 'in_image')]),
            (atlas_stats_node, self.output_node, [('atlas_statistics', 'atlas_statistics')])
        ])



# coding: utf-8

"""
 Module for converting PET_AV45 of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_av45_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert PET AV45 of ADNI

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the BIDS directory
        subjs_list: subject list

    """
    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of AV45 PET images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_av45_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of AV45 PET images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 'av45')
    cprint(Fore.GREEN + 'AV45 PET conversion done.' + Fore.RESET)


def compute_av45_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to the all PET images and store them in a tsv file

    Args:
        source_dir: path to the original ADNI folder
        csv_dir: path to clinical data folder
        dest_dir: path to the BIDS folder
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS

    """
    import pandas as pd
    import os
    import operator
    from os import walk, path
    from numpy import argsort
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint
    from functools import reduce

    pet_av45_col = [
            'Subject_ID', 'VISCODE', 'Visit',
            'Sequence', 'Scan_Date', 'Study_ID',
            'Series_ID', 'Image_ID', 'Original'
            ]

    pet_av45_df = pd.DataFrame(columns=pet_av45_col)
    av45qc_path = path.join(csv_dir, 'AV45QC.csv')
    pet_meta_list_path = path.join(csv_dir, 'PET_META_LIST.csv')
    petqc = pd.io.parsers.read_csv(av45qc_path, sep=',')
    pet_meta_list = pd.io.parsers.read_csv(pet_meta_list_path, sep=',')

    for subj in subjs_list:
        pet_qc_subj = petqc[(petqc.PASS == 1) & (petqc.RID == int(subj[-4:]))]
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]
        if subject_pet_meta.shape[0] < 1:
            # TODO Log somewhere subjects with problems
            cprint('NO Screening: Subject - ' + subj)
            continue

        for visit in list(pet_qc_subj.VISCODE2.unique()):
            pet_qc_visit = pet_qc_subj[pet_qc_subj.VISCODE2 == visit]
            if pet_qc_visit.shape[0] > 1:
                normal_images = []
                normal_meta = []
                for row in pet_qc_visit.iterrows():
                    image = row[1]
                    pet_meta_image = \
                        subject_pet_meta[(subject_pet_meta['Image ID'] == int(image.LONIUID[1:]))].iloc[0]
                    if pet_meta_image.Sequence.lower().find('early') < 0:
                        normal_images.append(image)
                        normal_meta.append(pet_meta_image)
                if len(normal_images) == 0:
                    # TODO Log somewhere subjects with problems
                    cprint('No regular AV45-PET image: Subject - ' + subj + ' for visit ' + visit)
                    continue
                if len(normal_images) == 1:
                    qc_visit = normal_images[0]
                else:
                    qc_visit = None
                    index = argsort([x['Series ID'] for x in normal_meta])
                    for i in index[::-1]:
                        coreg_avg = subject_pet_meta[(subject_pet_meta['Sequence'] == 'AV45 Co-registered, Averaged')
                                                     & (
                                                         subject_pet_meta['Series ID'] == normal_meta[i][
                                                             'Series ID'])]
                        if coreg_avg.shape[0] > 0:
                            qc_visit = normal_images[i]
                            break
                    if qc_visit is None:
                        qc_visit = normal_images[index[len(index) - 1]]
            else:
                qc_visit = pet_qc_visit.iloc[0]
            int_image_id = int(qc_visit.LONIUID[1:])
            original_pet_meta = subject_pet_meta[
                (subject_pet_meta['Orig/Proc'] == 'Original') & (subject_pet_meta['Image ID'] == int_image_id)
                & (subject_pet_meta.Sequence.map(lambda s: (s.lower().find('early') < 0)))]
            if original_pet_meta.shape[0] < 1:
                original_pet_meta = subject_pet_meta[
                        (subject_pet_meta['Orig/Proc'] == 'Original')
                        & (subject_pet_meta.Sequence.map(
                            lambda x: (x.lower().find('av45') > -1) & (x.lower().find('early') < 0))
                           )
                        & (subject_pet_meta['Scan Date'] == qc_visit.EXAMDATE)]
                if original_pet_meta.shape[0] < 1:
                    # TODO Log somewhere subjects with problems
                    cprint('NO Screening: Subject - ' + subj + ' for visit ' + qc_visit.VISCODE2)
                    continue
            original_image = original_pet_meta.iloc[0]
            averaged_pet_meta = subject_pet_meta[(subject_pet_meta['Sequence'] == 'AV45 Co-registered, Averaged') & (
                subject_pet_meta['Series ID'] == original_image['Series ID'])]
            if averaged_pet_meta.shape[0] < 1:
                sel_image = original_image
                original = True
            else:
                sel_image = averaged_pet_meta.iloc[0]
                original = False
            visit = sel_image.Visit
            sequence = replace_sequence_chars(sel_image.Sequence)
            date = sel_image['Scan Date']
            study_id = sel_image['Study ID']
            series_id = sel_image['Series ID']
            image_id = sel_image['Image ID']

            row_to_append = pd.DataFrame(
                [[subj, qc_visit.VISCODE2, str(visit), sequence, date, str(study_id), str(series_id), str(image_id),
                  original]],
                columns=pet_av45_col)
            pet_av45_df = pet_av45_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1
                         ('128_S_2220', 'm48')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((pet_av45_df.Subject_ID == conv_error[0])
                             & (pet_av45_df.VISCODE == conv_error[1]))

    indices_to_remove = pet_av45_df.index[reduce(operator.or_, error_indices, False)]
    pet_av45_df.drop(indices_to_remove, inplace=True)

    images = pet_av45_df
    # count = 0
    # total = images.shape[0]
    is_dicom = []
    image_folders = []
    for row in images.iterrows():
        image = row[1]
        seq_path = path.join(source_dir, str(image.Subject_ID), image.Sequence)
        # count += 1
        # print 'Processing Subject ' + str(image.Subject_ID) + ' - Session ' + image.VISCODE + ', ' + str(
        #     count) + ' / ' + str(total)
        image_path = ''
        for (dirpath, dirnames, filenames) in walk(seq_path):
            found = False
            for d in dirnames:
                if d == 'I' + str(image.Image_ID):
                    image_path = path.join(dirpath, d)
                    found = True
                    break
            if found:
                break

        dicom = True
        for (dirpath, dirnames, filenames) in walk(image_path):
            for f in filenames:
                if f.endswith(".nii"):
                    dicom = False
                    image_path = path.join(dirpath, f)
                    break

        is_dicom.append(dicom)
        image_folders.append(image_path)
        if image_path == '':
            cprint('No AV45 image path found for subject ' + str(image.Subject_ID))

    images.loc[:, 'Is_Dicom'] = pd.Series(is_dicom, index=images.index)
    images.loc[:, 'Path'] = pd.Series(image_folders, index=images.index)

    av45_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(av45_csv_path):
        os.mkdir(av45_csv_path)
    images.to_csv(path.join(av45_csv_path, 'av45_pet_paths.tsv'), sep='\t', index=False)
    return images



# coding: utf-8

"""
 Module for converting fMRI of ADNI
"""

__author__ = "Jorge Samper-Gonzalez and Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_fmri(source_dir, csv_dir, dest_dir, subjs_list=None):
    """

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination directory
        subjs_list: subjects list

    Returns:

    """
    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of fMRI images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    if path.isfile(path.join(dest_dir, 'conversion_info', 'fmri_paths.tsv')):
        images = pd.io.parsers.read_csv(path.join(dest_dir, 'conversion_info', 'fmri_paths.tsv'), sep='\t')
    else:

        images = compute_fmri_path(source_dir, csv_dir, dest_dir, subjs_list)

    cprint('Paths of fMRI images found. Exporting images into BIDS ...')

    fmri_paths_to_bids(dest_dir, images)
    cprint(Fore.GREEN + 'fMRI conversion done.' + Fore.RESET)


def compute_fmri_path(source_dir, clinical_dir, dest_dir, subjs_list):
    """
    Compute the paths to fmri images.

    The fmri images to convert into BIDS are chosen in the following way:
        - Extract the list of subjects from MAYOADIRL_MRI_FMRI_09_15_16.csv
        - Select the only the scans that came from PHILIPS machine (field Scanner from IDA_MR_Metadata_Listing.csv)
        - Discard all the subjects with column  series_quality = 4  (4 means that the scan is not usable) in MAYOADIRL_MRI_IMAGEQC_12_08_15.csv

    In case of multiple scans for the same session, same date the one to convert is chosen with the following criteria:
        - Check if in the file MAYOADIRL_MRI_IMAGEQC_12_08_15.csv there is a single scan with the field series_selected == 1
        - If yes choose the one with series_selected == 1
        - If no choose the scan with the best quality

    Args:
        source_dir: path to the ADNI image folder
        clinical_dir: path to the directory with all the clinical data od ADNI
        dest_dir:  path to the output_folder
        subjs_list: subjects list

    Returns: pandas Dataframe containing the path for each fmri

    """
    import os
    from os import path
    from os import walk
    import pandas as pd
    import logging
    from clinica.iotools.converters.adni_to_bids import adni_utils
    from clinica.utils.stream import cprint

    fmri_col = ['Subject_ID', 'VISCODE', 'Visit', 'IMAGEUID', 'Sequence', 'Scan Date', 'LONIUID', 'Scanner',
                'MagStregth', 'Path']

    fmri_df = pd.DataFrame(columns=fmri_col)

    # Load the requested clinical data
    mayo_mri_fmri_path = path.join(clinical_dir, 'MAYOADIRL_MRI_FMRI_09_15_16.csv')
    mayo_mri_imageqc_path = path.join(clinical_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv')
    ida_mr_metadata_path = path.join(clinical_dir, 'IDA_MR_Metadata_Listing.csv')

    mayo_mri_fmri = pd.io.parsers.read_csv(mayo_mri_fmri_path, sep=',')
    ida_mr_metadata = pd.io.parsers.read_csv(ida_mr_metadata_path, sep=',')
    mayo_mri_imageqc = pd.io.parsers.read_csv(mayo_mri_imageqc_path, sep=',')

    for subj in subjs_list:
        # print subj
        fmri_subjs_info = mayo_mri_fmri[(mayo_mri_fmri.RID == int(subj[-4:]))]
        # Extract visits available
        visits_list = fmri_subjs_info['VISCODE2'].tolist()
        # Removing duplicates
        visits_list = list(set(visits_list))

        if len(visits_list) != 0:
            for viscode in visits_list:
                visit = ''
                image_path = ''

                fmri_subj = fmri_subjs_info[fmri_subjs_info['VISCODE2'] == viscode]

                if not fmri_subj.empty:

                    # If there are multiple scans for the same session same subject, check what is the one selected for the usage (field 'series_selected') or
                    # choose the one with the best quality
                    if len(fmri_subj) > 1:
                        fmri_imageuid = fmri_subj['IMAGEUID'].tolist()
                        loni_uid_list = ['I' + str(imageuid) for imageuid in fmri_imageuid]
                        images_qc = mayo_mri_imageqc[mayo_mri_imageqc.loni_image.isin(loni_uid_list)]
                        series_selected_values = images_qc['series_selected'].tolist()
                        sum_series_selected = sum(series_selected_values)
                        if sum_series_selected == 1:
                            imageuid_to_select = images_qc[images_qc['series_selected'] > 0]['loni_image'].iloc[
                                0].replace('I', '')
                        else:
                            imageuid_to_select = select_image_qc(fmri_imageuid, images_qc)

                        fmri_subj = fmri_subj[fmri_subj['IMAGEUID'] == int(imageuid_to_select)].iloc[0]
                    else:
                        fmri_subj = fmri_subj.iloc[0]

                    fmri_imageuid = fmri_subj['IMAGEUID']

                    # Discard scans made with non Philips scanner and with a bad quality
                    fmri_metadata = ida_mr_metadata[ida_mr_metadata['IMAGEUID'] == fmri_imageuid]

                    if not fmri_metadata.empty:
                        fmri_metadata = fmri_metadata.iloc[0]

                        if 'Philips' not in fmri_metadata['Scanner']:
                            cprint('No Philips scanner for ' + subj + ' visit ' + viscode + '. Skipped.')
                            continue

                        elif 4 in mayo_mri_imageqc[mayo_mri_imageqc['loni_image'] == 'I' + str(fmri_imageuid)]['series_quality'].values:
                            cprint('Bad scan quality for ' + subj + ' visit ' + viscode + '. Skipped.')
                            continue

                        scan_date = fmri_subj.SCANDATE
                        sequence = adni_utils.replace_sequence_chars(fmri_subj.SERDESC)
                        scanner = fmri_metadata['Scanner']
                        loni_uid = fmri_metadata['LONIUID']
                        visit = fmri_metadata['Visit']
                        mag_strenght = fmri_metadata['MagStrength']

                        # Calculate the path
                        seq_path = path.join(source_dir, str(subj), sequence)
                        for (dirpath, dirnames, filenames) in walk(seq_path):
                            found = False
                            for d in dirnames:
                                if d == 'S' + str(loni_uid):
                                    image_path = path.join(dirpath, d)
                                    # Check if the path exists
                                    if not os.path.isdir(image_path):
                                        cprint('Path not existing for subject ' + subj + ' visit ' + visit)
                                    found = True
                                    break
                            if found:
                                break

                        # The session scmri correspond to the baseline
                        if viscode == 'scmri':
                            viscode = 'bl'
                    else:
                        cprint('Missing visit, sequence, scan date and loniuid for subject ' + subj + ' visit ' + visit)
                        continue

                    row_to_append = pd.DataFrame(
                        [[subj, str(viscode), visit, str(fmri_imageuid), sequence, scan_date, str(loni_uid),
                          scanner, mag_strenght, image_path]], columns=fmri_col)

                    fmri_df = fmri_df.append(row_to_append, ignore_index=True)
                else:
                    logging.info('Missing fMRI for ', subj, 'visit', visit)

    fmri_df.to_csv(path.join(dest_dir, 'conversion_info', 'fmri_paths.tsv'), sep='\t', index=False)
    return fmri_df


def fmri_paths_to_bids(dest_dir, fmri_paths, mod_to_update=False):
    """
    Convert the fmri extracted from the fmri_paths to BIDS

    Args:
        dest_dir: path to the input directory
        fmri_paths: path to the BIDS directory
        mod_to_add: if True add the fmri only where is missing
        mod_to_update:  if True overwrite (or create if is missing) all the existing fmri

    """
    from multiprocessing import cpu_count, Pool, Value
    from functools import partial

    counter = None

    def init(args):
        ''' store the counter for later use '''
        global counter
        counter = args

    subjs_list = fmri_paths['Subject_ID'].drop_duplicates().values

    counter = Value('i', 0)
    partial_generate_subject_files = partial(generate_subject_files,
                                             fmri_paths=fmri_paths,
                                             dest_dir=dest_dir,
                                             mod_to_update=mod_to_update)
    poolrunner = Pool(cpu_count(), initializer=init, initargs=(counter,))
    poolrunner.map(partial_generate_subject_files, subjs_list)
    del counter


def generate_subject_files(subj, fmri_paths, dest_dir, mod_to_update):
    import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils
    import clinica.iotools.bids_utils as bids_utils
    from clinica.utils.stream import cprint
    import subprocess
    import os
    import shutil
    from os import path
    from glob import glob

    sess_list = fmri_paths[(fmri_paths['Subject_ID'] == subj)]['VISCODE'].values
    alpha_id = adni_utils.remove_space_and_symbols(subj)
    bids_id = 'sub-ADNI' + alpha_id

    # For each session available, create the folder if doesn't exist and convert the files
    for ses in sess_list:
        with counter.get_lock():
            counter.value += 1
        cprint('[FLAIR] Processing subject ' + str(subj)
               + ' - session ' + ses + ', ' + str(counter.value)
               + ' / ' + str(len(fmri_paths)))
        ses_bids = adni_utils.viscode_to_session(ses)
        bids_ses_id = 'ses-' + ses_bids
        bids_file_name = bids_id + '_ses-' + ses_bids
        ses_path = path.join(dest_dir, bids_id, bids_ses_id)

        # If the fmri already exist
        existing_fmri = glob(path.join(ses_path, 'func', '*_bold*'))
        # if mod_to_add:
        #     if len(existing_fmri) > 0:
        #         print 'Fmri already existing. Skipped.'
        #         continue

        if mod_to_update and len(existing_fmri) > 0:
            # print 'Removing the old fmri folder...'
            os.remove(existing_fmri[0])

        if not os.path.exists(ses_path):
            if not os.path.exists(path.join(dest_dir, bids_id)):
                os.mkdir(path.join(dest_dir, bids_id))
            os.mkdir(path.join(dest_dir, bids_id, bids_ses_id))

        fmri_info = fmri_paths[(fmri_paths['Subject_ID'] == subj) & (fmri_paths['VISCODE'] == ses)]
        if not fmri_info['Path'].values[0] == '':
            if type(fmri_info['Path'].values[0]) != float:
                if not os.path.exists(path.join(ses_path, 'func')):
                    os.mkdir(path.join(ses_path, 'func'))
                fmri_path = fmri_info['Path'].values[0]
                dcm_to_convert = adni_utils.check_two_dcm_folder(fmri_path,
                                                                 dest_dir,
                                                                 fmri_info['IMAGEUID'].values[0])
                if not os.path.isfile(os.path.join(ses_path, 'func', bids_file_name + '_task-rest_bold.nii.gz')):
                    bids_utils.convert_fmri(dcm_to_convert, path.join(ses_path, 'func'), bids_file_name)
                else:
                    cprint("Images already converted")

                # Delete the temporary folder used for copying fmri with 2 subjects inside the DICOM folder
                adni_utils.remove_tmp_dmc_folder(dest_dir, fmri_info['IMAGEUID'].values[0])


def select_image_qc(id_list, mri_qc_subj):
    """

    Args:
        id_list:
        mri_qc_subj:

    Returns:

    """
    import numpy as np

    if len(id_list) == 0:
        return None

    selected_image = None
    image_ids = ['I' + str(imageuid) for imageuid in id_list]
    int_ids = [int(imageuid) for imageuid in id_list]
    images_qc = mri_qc_subj[mri_qc_subj.loni_image.isin(image_ids)]

    if images_qc.shape[0] < 1:
        return min(int_ids)

    if np.sum(images_qc.series_selected) == 1:
        selected_image = images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
    else:
        images_not_rejected = images_qc[images_qc.series_quality < 4]

        if images_not_rejected.shape[0] < 1:

            # There are no images that passed the qc
            # so we'll try to see if there are other images without qc,
            # otherwise return None
            qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
            no_qc_ids = list(set(int_ids) - qc_ids)

            if len(no_qc_ids) == 0:
                return None
            else:
                return min(no_qc_ids)

        series_quality = [q if q > 0 else 4 for q in list(images_not_rejected.series_quality)]
        best_q = np.amin(series_quality)

        images_best_qc = images_not_rejected[images_not_rejected.series_quality == best_q]
        if images_best_qc.shape[0] == 1:
            selected_image = images_best_qc.iloc[0].loni_image[1:]
        else:
            selected_image = min(int_ids)

    return int(selected_image)


def compute_fmri_path_refactoring(source_dir, clinical_dir, dest_dir, subjs_list):
    """
    Compute the paths to fmri images.

    The fmri images to convert into BIDS are chosen in the following way:
        - Extract the list of subjects from MAYOADIRL_MRI_FMRI_09_15_16.csv
        - Select the only the scans that came from PHILIPS machine (field Scanner from IDA_MR_Metadata_Listing.csv)
        - Discard all the subjects with column  series_quality = 4  (4 means that the scan is not usable) in MAYOADIRL_MRI_IMAGEQC_12_08_15.csv

    In case of multiple scans for the same session, same date the one to convert is chosen with the following criteria:
        - Check if in the file MAYOADIRL_MRI_IMAGEQC_12_08_15.csv there is a single scan with the field series_selected == 1
        - If yes choose the one with series_selected == 1
        - If no choose the scan with the best quality

    Args:
        source_dir: path to the ADNI image folder
        clinical_dir: path to the directory with all the clinical data od ADNI
        dest_dir:  path to the output_folder
        subjs_list: subjects list

    Returns: pandas Dataframe containing the path for each fmri

    """
    import os
    from os import path
    from os import walk
    import pandas as pd
    import logging
    from clinica.iotools.converters.adni_to_bids import adni_utils
    from clinica.utils.stream import cprint

    fmri_col = ['Subject_ID', 'VISCODE', 'Visit', 'IMAGEUID', 'Sequence', 'Scan Date', 'LONIUID', 'Scanner',
                'MagStregth', 'Path']

    fmri_df = pd.DataFrame(columns=fmri_col)

    # Load the requested clinical data
    mayo_mri_fmri_path = path.join(clinical_dir, 'MAYOADIRL_MRI_FMRI_09_15_16.csv')
    mayo_mri_imageqc_path = path.join(clinical_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv')

    mayo_mri_fmri = pd.io.parsers.read_csv(mayo_mri_fmri_path, sep=',')
    mayo_mri_imageqc = pd.io.parsers.read_csv(mayo_mri_imageqc_path, sep=',')

    for subj in subjs_list:
        # print subj
        fmri_subjs_info = mayo_mri_fmri[(mayo_mri_fmri.RID == int(subj[-4:]))]
        # Extract visits available
        visits_list = fmri_subjs_info['VISCODE2'].tolist()
        # Removing duplicates
        visits_list = list(set(visits_list))

        if len(visits_list) != 0:
            for viscode in visits_list:
                visit = ''
                image_path = ''

                fmri_subj = fmri_subjs_info[fmri_subjs_info['VISCODE2'] == viscode]

                if not fmri_subj.empty:

                    # If there are multiple scans for the same session same subject, check what is the one selected for the usage (field 'series_selected') or
                    # choose the one with the best quality
                    if len(fmri_subj) > 1:
                        fmri_imageuid = fmri_subj['IMAGEUID'].tolist()
                        loni_uid_list = ['I' + str(imageuid) for imageuid in fmri_imageuid]
                        images_qc = mayo_mri_imageqc[mayo_mri_imageqc.loni_image.isin(loni_uid_list)]
                        series_selected_values = images_qc['series_selected'].tolist()
                        sum_series_selected = sum(series_selected_values)
                        if sum_series_selected == 1:
                            imageuid_to_select = images_qc[images_qc['series_selected'] > 0]['loni_image'].iloc[
                                0].replace('I', '')
                        else:
                            imageuid_to_select = select_image_qc(fmri_imageuid, images_qc)

                        fmri_subj = fmri_subj[fmri_subj['IMAGEUID'] == int(imageuid_to_select)].iloc[0]
                    else:
                        fmri_subj = fmri_subj.iloc[0]

                    fmri_imageuid = fmri_subj['IMAGEUID']

                    # Discard scans made with non Philips scanner and with a bad quality
                    fmri_metadata = ida_mr_metadata[ida_mr_metadata['IMAGEUID'] == fmri_imageuid]

                    if not fmri_metadata.empty:
                        fmri_metadata = fmri_metadata.iloc[0]

                        if 'Philips' not in fmri_metadata['Scanner']:
                            cprint('No Philips scanner for ' + subj + ' visit ' + viscode + '. Skipped.')
                            continue

                        elif 4 in mayo_mri_imageqc[mayo_mri_imageqc['loni_image'] == 'I' + str(fmri_imageuid)]['series_quality'].values:
                            cprint('Bad scan quality for ' + subj + ' visit ' + viscode + '. Skipped.')
                            continue

                        scan_date = fmri_subj.SCANDATE
                        sequence = adni_utils.replace_sequence_chars(fmri_subj.SERDESC)
                        scanner = fmri_metadata['Scanner']
                        loni_uid = fmri_metadata['LONIUID']
                        visit = fmri_metadata['Visit']
                        mag_strenght = fmri_metadata['MagStrength']

                        # Calculate the path
                        seq_path = path.join(source_dir, str(subj), sequence)
                        for (dirpath, dirnames, filenames) in walk(seq_path):
                            found = False
                            for d in dirnames:
                                if d == 'S' + str(loni_uid):
                                    image_path = path.join(dirpath, d)
                                    # Check if the path exists
                                    if not os.path.isdir(image_path):
                                        cprint('Path not existing for subject ' + subj + ' visit ' + visit)
                                    found = True
                                    break
                            if found:
                                break

                        # The session scmri correspond to the baseline
                        if viscode == 'scmri':
                            viscode = 'bl'
                    else:
                        cprint('Missing visit, sequence, scan date and loniuid for subject ' + subj + ' visit ' + visit)
                        continue

                    row_to_append = pd.DataFrame(
                        [[subj, str(viscode), visit, str(fmri_imageuid), sequence, scan_date, str(loni_uid),
                          scanner, mag_strenght, image_path]], columns=fmri_col)

                    fmri_df = fmri_df.append(row_to_append, ignore_index=True)
                else:
                    logging.info('Missing fMRI for ', subj, 'visit', visit)

    fmri_df.to_csv(path.join(dest_dir, 'conversion_info', 'fmri_paths.tsv'), sep='\t', index=False)
    return fmri_df



# coding: utf8

import os
from nipype.interfaces.spm.base import SPMCommandInputSpec, SPMCommand, scans_for_fnames
from nipype.interfaces.base import TraitedSpec, File, traits
from nipype.utils.filemanip import split_filename

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


if 'SPMSTANDALONE_HOME' in os.environ:
    if 'MCR_HOME' in os.environ:
        matlab_cmd = os.environ['SPMSTANDALONE_HOME'] \
                + 'run_spm12.sh ' \
                + os.environ['MCR_HOME'] \
                + ' script'
        SPMCommand.set_mlab_paths(matlab_cmd=matlab_cmd, use_mcr=True)


class DARTELExistingTemplateInputSpec(SPMCommandInputSpec):
    image_files = traits.List(traits.List(File(exists=True)),
                              desc="A list of files to be segmented",
                              field='warp1.images', copyfile=False, mandatory=True)
    regularization_form = traits.Enum('Linear', 'Membrane', 'Bending',
                                      field='warp1.settings.rform',
                                      desc='Form of regularization energy term')
    iteration_parameters = traits.List(traits.Tuple(traits.Range(1, 10),
                                                    traits.Tuple(traits.Float,
                                                                 traits.Float,
                                                                 traits.Float),
                                                    traits.Range(0, 9),
                                                    traits.File(exists=True)),
                                       minlen=3,
                                       maxlen=12,
                                       mandatory=True,
                                       field='warp1.settings.param',
                                       desc="""List of tuples for each iteration
                                       - Inner iterations
                                       - Regularization parameters
                                       - Time points for deformation model
                                       - DARTEL template
                                       """)
    optimization_parameters = traits.Tuple(traits.Float, traits.Range(1, 8),
                                           traits.Range(1, 8),
                                           field='warp1.settings.optim',
                                           desc="""Optimization settings a tuple
                                           - LM regularization
                                           - cycles of multigrid solver
                                           - relaxation iterations
                                           """)


class DARTELExistingTemplateOutputSpec(TraitedSpec):
    dartel_flow_fields = traits.List(File(exists=True), desc='DARTEL flow fields')


class DARTELExistingTemplate(SPMCommand):
    """Use spm DARTEL to create a template and flow fields
    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=185
    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> dartel = spm.DARTEL()
    >>> dartel.inputs.image_files = [['rc1s1.nii','rc1s2.nii'],['rc2s1.nii', 'rc2s2.nii']]
    >>> dartel.run() # doctest: +SKIP
    """

    input_spec = DARTELExistingTemplateInputSpec
    output_spec = DARTELExistingTemplateOutputSpec
    _jobtype = 'tools'
    _jobname = 'dartel'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        import numpy as np

        if opt in ['image_files']:
            return scans_for_fnames(val, keep4d=True, separate_sessions=True)
        elif opt == 'regularization_form':
            mapper = {'Linear': 0, 'Membrane': 1, 'Bending': 2}
            return mapper[val]
        elif opt == 'iteration_parameters':
            params = []
            for param in val:
                new_param = {}
                new_param['its'] = param[0]
                new_param['rparam'] = list(param[1])
                new_param['K'] = param[2]
                new_param['template'] = np.array([param[3]], dtype=object)
                params.append(new_param)
            return params
        elif opt == 'optimization_parameters':
            new_param = {}
            new_param['lmreg'] = val[0]
            new_param['cyc'] = val[1]
            new_param['its'] = val[2]
            return [new_param]
        else:
            return super(DARTELExistingTemplate, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['dartel_flow_fields'] = []
        for filename in self.inputs.image_files[0]:
            pth, base, ext = split_filename(filename)
            outputs['dartel_flow_fields'].append(os.path.realpath('u_%s%s' % (base, ext)))
        return outputs


def prepare_dartel_input_images(dartel_input_images):
    return [[[tissue] for tissue in subject] for subject in zip(*dartel_input_images)]


def prepare_images_from_segmentation(dartel_input_images):
    return [[tissue for tissue in subject if tissue] for subject in dartel_input_images if subject]


def create_iteration_parameters(dartel_templates, iteration_parameters):

    if len(dartel_templates) != 6:
        raise ValueError('Wrong number of templates. 6 templates were expected, one for each DARTEL iteration.')

    if iteration_parameters is None:
        iter1 = (3, (4, 2, 1e-06), 0, dartel_templates[0])
        iter2 = (3, (2, 1, 1e-06), 0, dartel_templates[1])
        iter3 = (3, (1, 0.5, 1e-06), 1, dartel_templates[2])
        iter4 = (3, (0.5, 0.25, 1e-06), 2, dartel_templates[3])
        iter5 = (3, (0.25, 0.125, 1e-06), 4, dartel_templates[4])
        iter6 = (3, (0.25, 0.125, 1e-06), 6, dartel_templates[5])

        return [iter1, iter2, iter3, iter4, iter5, iter6]

    elif len(iteration_parameters) != 6:
        raise ValueError('Wrong number of iteration parameters. 6 iterations were expected.')
    else:
        new_iteration_parameters = []
        for i in range(6):
            new_iteration_parameters.append((iteration_parameters[i][0],
                                             iteration_parameters[i][1],
                                             iteration_parameters[i][2],
                                             dartel_templates[i])
                                            )
        return new_iteration_parameters


def get_class_images(class_images, index_list):
    """
    utility method to extract class images
    from a multi session class_images set:
    class nb : tissue type
    1 : Grey Matter
    2 : White Matter
    3 : CerebroSpinal Fluid
    4 : Skull
    5 : Out-of-brain soft tissue
    6 : Head surrounding

    :param class_images: image set from which to extract images.
    :param index_list: index list of the classes to extract.
    :return: extracted images in a list of lists (without empty lists).

    Example
    -------
    >>> class_n_images = get_class_images(class_images,[1,2])
    """

    # Declare class images list
    class_n_images = {}
    for idx in index_list:
        class_n_images[idx] = []

    for session in class_images:
        for idx in index_list:
            class_n_images[idx].extend(session[idx-1])

    result = []
    for idx in index_list:
        if class_n_images[idx]:
            result.append([class_n_images[idx]])

    return result



# coding: utf8


"""This module contains FSL utilities."""


def standard_space_roi(in_t1, in_mask=None):
    """
    Pre-mask the structural image to standard space.

    The purpose of the FSL command standard_space_roi is to do an initial
    masking of the image before performing a brain extraction with FSL BET.
    This uses flirt to register the input image to a standard space whole-head
    image; the resulting transform is then inverted and a standard-space brain
    mask is transformed into the space of the input image, and then applied to
    this to create the output.

    Args:
        in_t1 (str): File containing the T1-weighted image.
        in_mask (Optional[str]): Mask output using transformed standard space
            mask (default=None, i.e. it will use the 2mm dilated MNI mask
            located in ${FSLDIR}/data/standard).

    Returns:
        out_pre_mask (str): Pre-mask the structural image.
    """
    import os.path as op
    import os

    assert(op.isfile(in_t1))

    out_pre_mask = op.abspath('T1_pre_bet.nii.gz')

    cmd = 'standard_space_roi %s %s -roiNONE' % (in_t1, out_pre_mask)
    if in_mask is None:
        cmd = cmd + ' -b'
    else:
        cmd = cmd + ' -maskMASK ' + in_mask
    os.system(cmd)
    return out_pre_mask



# coding: utf8

import clinica.engine as ce


class DwiPreprocessingUsingT1Cli(ce.CmdParser):

    def __init__(self):
        super(DwiPreprocessingUsingT1Cli, self).__init__()

    def define_name(self):
        """Define the sub-command name to run this pipeline.
        """
        self._name = 'dwi-preprocessing-using-t1'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Preprocessing of raw DWI datasets using a T1w image:\nhttp://clinica.run/doc/Pipelines/DWI_Preprocessing/'

    def define_options(self):
        """Define the sub-command arguments.
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("--low_bval",
                              metavar=('N'), type=int, default=5,
                              help='Define the b0 volumes as all volume bval <= lowbval. (default: --low_bval 5)')  # noqa
        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-tsv", "--subjects_sessions_tsv",
                                 help='TSV file containing a list of subjects with their sessions.')
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results')
        clinica_opt.add_argument("-np", "--n_procs",
                                 metavar=('N'), type=int,
                                 help='Number of cores used to run in parallel')

    def run_command(self, args):
        """
        Run the DWIPreprocessingUsingT1 Pipeline from command line.
        """
        from tempfile import mkdtemp
        from clinica.utils.stream import cprint
        from .dwi_preprocessing_using_t1_pipeline import DwiPreprocessingUsingT1

        pipeline = DwiPreprocessingUsingT1(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            low_bval=args.low_bval
        )

        if args.working_directory is None:
            args.working_directory = mkdtemp()
        pipeline.base_dir = self.absolute_path(args.working_directory)

        if args.n_procs:
            pipeline.run(plugin='MultiProc',
                         plugin_args={'n_procs': args.n_procs})
        else:
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8


"""This module contains MRtrix utilities."""


def dilate_mask(in_mask, npass=4, nthreads=2):
    """
    Dilate the mask.

    This function performs the dilation of a given binary (brain) mask.

    Args:
        in_mask (str): Binary mask.
        npass (Optional[int]): Number of times to repeatedly apply the dilation
            (default=4).
        nthreads (Optional[int]): Number of threads used in this function
            (default=2, 0 disables multi-threading).

    Returns:
        The dilated mask.
    """
    import os.path as op
    import os

    assert(op.isfile(in_mask))

    out_dilated_mask = op.abspath('dilated_mask.nii')
    cmd = 'maskfilter -npass %s -nthreads %s %s dilate %s' % \
          (npass, nthreads, in_mask, out_dilated_mask)
    os.system(cmd)

    return out_dilated_mask



# coding: utf8

from os.path import dirname
from os.path import realpath
from os.path import join as pathjoin

from grabbit import Layout

__all__ = ['CAPSLayout']


class CAPSLayout(Layout):

    def __init__(self, path, config=None, **kwargs):
        import os
        if config is None:
            root = dirname(realpath(__file__))
            config = pathjoin(os.path.abspath(root), '..', 'resources', 'layouts', 'caps.json')
        super(CAPSLayout, self).__init__(path, config,
                                         dynamic_getters=True, **kwargs)


# if __name__ == '__main__':
#     layout = CAPSLayout(path='/Users/jeremy.guillon/Repositories/multiconproject/data/HMTC_CAPS')
#     #print(layout.get(target='subject', return_type='dir', regex_search=True))
#     print(layout.get(target='session', return_type='dir', regex_search=False))
#     print(layout.get(fmripreprocessing='bold_motionparams'))
#     print(layout.get(target='modality', return_type='dir', regex_search=False))
#     print(layout.get(target='modality', return_type='id', regex_search=False))



# coding: utf8

import clinica.engine as ce


class T1VolumeDartel2MNICLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 't1-volume-dartel2mni'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Register DARTEL template to MNI space:\n'\
                            + 'http://clinica.run/doc/Pipelines/T1_Volume/'

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-s", "--smooth",
                              nargs='+', type=int, default=[8],
                              help="A list of integers specifying the different isomorphic FWHM in millimeters to smooth the image (default: --smooth 8).")
        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-tsv", "--subjects_sessions_tsv",
                                 help='TSV file containing a list of subjects with their sessions.')
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results')
        clinica_opt.add_argument("-np", "--n_procs",
                                 metavar=('N'), type=int,
                                 help='Number of cores used to run in parallel')
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-t", "--tissues", nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help='Tissues to create flow fields to DARTEL template (default: GM, WM and CSF i.e. --tissues 1 2 3).')
        advanced.add_argument("-m", "--modulate",
                              type=bool, default=True,
                              metavar=('True/False'),
                              help='A boolean. Modulate output images - no modulation preserves concentrations (default: --modulate True).')
        advanced.add_argument("-vs", "--voxel_size",
                              metavar=('float'),
                              nargs=3, type=float,
                              help="A list of 3 floats specifying the voxel size of the output image (default: --voxel_size 1.5 1.5 1.5).")
        list_atlases = ['AAL2', 'LPBA40', 'Neuromorphometrics', 'AICHA', 'Hammers']
        advanced.add_argument("-atlases", "--atlases",
                              nargs='+', type=str, metavar='',
                              default=list_atlases, choices=list_atlases,
                              help='A list of atlases used to calculate the regional mean GM concentrations (default: all atlases i.e. --atlases AAL2 AICHA Hammers LPBA40 Neuromorphometrics).')

    def run_command(self, args):
        """
        """
        from tempfile import mkdtemp
        from clinica.utils.stream import cprint
        from clinica.pipelines.t1_volume_dartel2mni.t1_volume_dartel2mni_pipeline import T1VolumeDartel2MNI

        pipeline = T1VolumeDartel2MNI(
                bids_directory=self.absolute_path(args.bids_directory),
                caps_directory=self.absolute_path(args.caps_directory),
                tsv_file=self.absolute_path(args.subjects_sessions_tsv),
                group_id=args.group_id
                )

        pipeline.parameters.update({
            'tissues': args.tissues,
            # 'bounding_box': None,
            'voxel_size': tuple(args.voxel_size) if args.voxel_size is not None else None,
            'modulation': args.modulate,
            'fwhm': args.smooth,
            'atlas_list': args.atlases
        })

        if args.working_directory is None:
            args.working_directory = mkdtemp()
        pipeline.base_dir = self.absolute_path(args.working_directory)

        if args.n_procs:
            pipeline.run(plugin='MultiProc', plugin_args={'n_procs': args.n_procs})
        else:
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def get_class_images(class_images, index_list):
    """
    utility method to extract class images
    from a multi session class_images set:
    class nb : tissue type
    1 : Grey Matter
    2 : White Matter
    3 : CerebroSpinal Fluid
    4 : Skull
    5 : Out-of-brain soft tissue
    6 : Head surrounding

    :param class_images: image set from which to extract images.
    :param index_list: index list of the classes to extract.
    :return: extracted images in a list of lists (without empty lists).

    Example
    -------
    >>> class_n_images = get_class_images(class_images,[1,2])
    """

    # Declare class images list
    class_n_images = {}
    for idx in index_list:
        class_n_images[idx] = []

    for session in class_images:
        for idx in index_list:
            class_n_images[idx].extend(session[idx-1])

    result = []
    for idx in index_list:
        if class_n_images[idx]:
            result.append(class_n_images[idx])

    return result



# coding: utf8

import clinica.pipelines.engine as cpe

__author__ = "Junhao Wen"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Junhao Wen", "Arnaud Marcoux", "Alexandre Routier"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Junhao Wen"
__email__ = "Junhao.Wen@inria.fr"
__status__ = "Development"


class StatisticsSurface(cpe.Pipeline):
    """
    Based on the Matlab toolbox [SurfStat](http://www.math.mcgill.ca/keith/surfstat/), which performs statistical
    analyses of univariate and multivariate surface and volumetric data using the generalized linear model (GLM),
    this pipelines performs analyses including group comparison and correlation with the surface-based features.
    Currently, this pipelines fits the normalised cortical thickness on FsAverage from `t1-freesurfer` pipelines.
    New features will be added in the future.


    Args:
        caps_directory: str, the output folder of recon-all which will contain the result files: ?h.thickness.fwhm**.mgh.
        tsv_file: str, Path to the tsv containing the information for GLM.
        design_matrix: str, the linear model that fits into the GLM, for example '1+group'.
        contrast: string, the contrast matrix for GLM, if the factor you choose is categorized variable, clinica_surfstat will create two contrasts,
                  for example, contrast = 'Label', this will create contrastpos = Label.AD - Label.CN, contrastneg = Label.CN - Label.AD; if the fac-
                  tory that you choose is a continuous factor, clinica_surfstat will just create one contrast, for example, contrast = 'Age', but note,
                  the string name that you choose should be exactly the same with the columns names in your subjects_visits_tsv.
        str_format: string, the str_format which uses to read your tsv file, the typy of the string should corresponds exactly with the columns in the tsv file.
            Defaut parameters, we set these parameters to be some default values, but you can also set it by yourself:
        group_label: current group name for this analysis
        glm_type: based on the hypothesis, you should define one of the glm types, "group_comparison", "correlation"
        full_width_at_half_maximum: fwhm for the surface smoothing, default is 20, integer.
        threshold_uncorrected_pvalue: threshold to display the uncorrected Pvalue, float, default is 0.001.
        threshold_corrected_pvalue: the threshold to display the corrected cluster, default is 0.05, float.
        cluster_threshold: threshold to define a cluster in the process of cluster-wise correction, default is 0.001, float.
        working_directory: define where to put the infomartion of the nipype workflow.
        n_procs: define how many cores to run this workflow.

    Returns:
        A clinica pipeline object containing the StatisticsSurface pipeline.

    """
    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['design_matrix',
                'contrast',
                'str_format',
                'group_label',
                'glm_type',
                'surface_file',
                'full_width_at_half_maximum',
                'threshold_uncorrected_pvalue',
                'threshold_corrected_pvalue',
                'cluster_threshold',
                'feature_label']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return []

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True))
        read_parameters_node.inputs.design_matrix = self.parameters['design_matrix']
        read_parameters_node.inputs.contrast = self.parameters['contrast']
        read_parameters_node.inputs.str_format = self.parameters['str_format']
        read_parameters_node.inputs.group_label = self.parameters['group_label']
        read_parameters_node.inputs.glm_type = self.parameters['glm_type']
        read_parameters_node.inputs.surface_file = self.parameters['custom_file']
        read_parameters_node.inputs.full_width_at_half_maximum = self.parameters['full_width_at_half_maximum']
        read_parameters_node.inputs.threshold_uncorrected_pvalue = self.parameters['threshold_uncorrected_pvalue']
        read_parameters_node.inputs.threshold_corrected_pvalue = self.parameters['threshold_corrected_pvalue']
        read_parameters_node.inputs.cluster_threshold = self.parameters['cluster_threshold']
        read_parameters_node.inputs.feature_label = self.parameters['feature_label']

        self.connect([
            (read_parameters_node, self.input_node, [('design_matrix',                'design_matrix')]),  # noqa
            (read_parameters_node, self.input_node, [('surface_file',                 'surface_file')]),  # noqa
            (read_parameters_node, self.input_node, [('contrast',                     'contrast')]),  # noqa
            (read_parameters_node, self.input_node, [('str_format',                   'str_format')]),  # noqa
            (read_parameters_node, self.input_node, [('group_label',                  'group_label')]),  # noqa
            (read_parameters_node, self.input_node, [('glm_type',                     'glm_type')]),  # noqa
            (read_parameters_node, self.input_node, [('full_width_at_half_maximum',   'full_width_at_half_maximum')]),  # noqa
            (read_parameters_node, self.input_node, [('threshold_uncorrected_pvalue', 'threshold_uncorrected_pvalue')]),  # noqa
            (read_parameters_node, self.input_node, [('threshold_corrected_pvalue',   'threshold_corrected_pvalue')]),  # noqa
            (read_parameters_node, self.input_node, [('cluster_threshold',            'cluster_threshold')]),  # noqa
            (read_parameters_node, self.input_node, [('feature_label',                'feature_label')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """

        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """

        import clinica.pipelines.statistics_surface.statistics_surface_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.io import JSONFileSink

        # Node to fetch the input variables
        data_prep = npe.Node(name='inputnode',
                             interface=nutil.Function(
                                 input_names=['input_directory', 'subjects_visits_tsv', 'group_label', 'glm_type'],
                                 output_names=['path_to_matscript', 'surfstat_input_dir', 'output_directory',
                                               'freesurfer_home', 'out_json'],
                                 function=utils.prepare_data))
        data_prep.inputs.input_directory = self.caps_directory
        data_prep.inputs.subjects_visits_tsv = self.tsv_file

        # Node to wrap the SurfStat matlab script
        surfstat = npe.Node(name='surfstat',
                            interface=nutil.Function(
                                input_names=['input_directory',
                                             'output_directory',
                                             'subjects_visits_tsv',
                                             'design_matrix',
                                             'contrast',
                                             'str_format',
                                             'glm_type',
                                             'group_label',
                                             'freesurfer_home',
                                             'surface_file',
                                             'path_to_matscript',
                                             'full_width_at_half_maximum',
                                             'threshold_uncorrected_pvalue',
                                             'threshold_corrected_pvalue',
                                             'cluster_threshold',
                                             'feature_label'],
                                output_names=['out_images'],
                                function=utils.runmatlab))
        surfstat.inputs.subjects_visits_tsv = self.tsv_file

        # Node to create the dictionary for JSONFileSink
        jsondict = npe.Node(name='Jsondict',
                            interface=nutil.Function(
                                input_names=['glm_type', 'design_matrix', 'str_format', 'contrast', 'group_label', 'full_width_at_half_maximum',
                                             'threshold_uncorrected_pvalue', 'threshold_corrected_pvalue', 'cluster_threshold'],
                                output_names=['json_dict'],
                                function=utils.json_dict_create))

        # Node to write the GLM information into a JSON file
        jsonsink = npe.Node(JSONFileSink(input_names=['out_file']), name='jsonsinker')

        # Connection
        # ==========
        self.connect([
            (self.input_node, data_prep, [('group_label', 'group_label')]),  # noqa
            (self.input_node, data_prep, [('glm_type', 'glm_type')]),  # noqa
            (self.input_node, surfstat, [('design_matrix', 'design_matrix')]),  # noqa
            (self.input_node, surfstat, [('contrast', 'contrast')]),  # noqa
            (self.input_node, surfstat, [('str_format', 'str_format')]),  # noqa
            (self.input_node, surfstat, [('glm_type', 'glm_type')]),  # noqa
            (self.input_node, surfstat, [('group_label', 'group_label')]),  # noqa
            (self.input_node, surfstat, [('full_width_at_half_maximum', 'full_width_at_half_maximum')]),  # noqa
            (self.input_node, surfstat, [('threshold_uncorrected_pvalue', 'threshold_uncorrected_pvalue')]),  # noqa
            (self.input_node, surfstat, [('threshold_corrected_pvalue', 'threshold_corrected_pvalue')]),  # noqa
            (self.input_node, surfstat, [('cluster_threshold', 'cluster_threshold')]),  # noqa
            (self.input_node, surfstat, [('surface_file', 'surface_file')]),  # noqa
            (self.input_node, surfstat, [('feature_label', 'feature_label')]),
            (data_prep, surfstat, [('surfstat_input_dir', 'input_directory')]),  # noqa
            (data_prep, surfstat, [('path_to_matscript', 'path_to_matscript')]),  # noqa
            (data_prep, surfstat, [('output_directory', 'output_directory')]),  # noqa
            (data_prep, surfstat, [('freesurfer_home', 'freesurfer_home')]),  # noqa
            (self.input_node, jsondict, [('glm_type', 'glm_type')]),  # noqa
            (self.input_node, jsondict, [('design_matrix', 'design_matrix')]),  # noqa
            (self.input_node, jsondict, [('str_format', 'str_format')]),  # noqa
            (self.input_node, jsondict, [('contrast', 'contrast')]),  # noqa
            (self.input_node, jsondict, [('group_label', 'group_label')]),  # noqa
            (self.input_node, jsondict, [('full_width_at_half_maximum', 'full_width_at_half_maximum')]),  # noqa
            (self.input_node, jsondict, [('threshold_uncorrected_pvalue', 'threshold_uncorrected_pvalue')]),  # noqa
            (self.input_node, jsondict, [('threshold_corrected_pvalue', 'threshold_corrected_pvalue')]),  # noqa
            (self.input_node, jsondict, [('cluster_threshold', 'cluster_threshold')]),  # noqa
            (data_prep, jsonsink, [('out_json', 'out_file')]),  # noqa
            (jsondict, jsonsink, [('json_dict', 'in_dict')]),  # noqa
        ])



# coding: utf8

# WARNING: Don't put any import statement here except if it's absolutly
# necessary. Put it *inside* the different methods.
# Otherwise it will slow down the dynamic loading of the pipelines list by the
# command line tool.
import clinica.pipelines.engine as cpe

__author__ = "Jeremy Guillon"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jeremy Guillon", "Romain Valabregue"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jeremy Guillon"
__email__ = "jeremy.guillon@inria.fr"
__status__ = "Development"


class fMRIPreprocessing(cpe.Pipeline):
    """Create fMRI preprocessing pipelines object.

    Warnings:
        - The Fieldmap node is still under revision as a pull request
        - The RealingUnwarp node is still under revision as a pull request

    Todos:
        - [x] Don't read inputs if not needed (i.e. --unwarp or no)
        - [x] Read parameters from sidecar `*.json` files.
        - [x] Add support of gzipped nifti inputs.
        - [x] Replace reg_node target image by the brain only using c1 + c2 +
        c3 dilated-eroded-filled.
        - [x] Develop SPM Realign and Unwarp wrapper and integrate it.
        - [x] Develop SPM Fieldmap Calculation Tool wrapper and integrate it.
        - [x] Replace standard DataGrabber by a BIDS tree finder.
        - [x] Export only gzipped nifti files.

    Args:
        input_dir: A BIDS directory.
        output_dir: An empty output directory where CAPS structured data will
        be written.
        subjects_sessions_list: The Subjects-Sessions list file (in .tsv
        format).

    Returns:
        A nipype workflow object containing the full fMRI preprocessing
        pipelines.

    Raises:
        IOError:

    Example:
        >>> from clinica.pipelines.fmri_preprocessing
        .fmri_preprocessing_pipeline import fMRIPreprocessing
        >>> pipelines = fMRIPreprocessing('~/MYDATASET_BIDS',
        '~/MYDATASET_CAPS')
        >>> pipelines.parameters = {
        >>>     'num_slices' : 45,
        >>>     'time_repetition' : 2.4,
        >>>     'echo_times' : [5.19, 7.65],
        >>>     'blip_direction' : 1,
        >>>     'total_readout_time' : 15.6799,
        >>>     'full_width_at_half_maximum' : [8, 8, 8],
        >>>     't1_native_space' : False
        >>> }
        >>> pipelines.run()
    """

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        if ('unwarping' in self.parameters) and self.parameters['unwarping']:
            return ['et', 'blipdir', 'tert', 'time_repetition', 'num_slices',
                    'magnitude1', 'slice_order', 'ref_slice',
                    'time_acquisition', 'phasediff', 'bold', 'T1w']
        else:
            return ['time_repetition', 'num_slices', 'slice_order', 'ref_slice',
                    'time_acquisition', 'bold', 'T1w']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        if ('t1_native_space' in self.parameters) and self.parameters[
                't1_native_space']:
            return ['t1_brain_mask', 'mc_params', 'native_fmri', 't1_fmri',
                    'mni_fmri', 'mni_smoothed_fmri']
        else:
            return ['t1_brain_mask', 'mc_params', 'native_fmri', 'mni_fmri',
                    'mni_smoothed_fmri']

    def build_input_node(self):
        """Build and connect an input node to the pipelines.

        References:
            https://lcni.uoregon.edu/kb-articles/kb-0003

        """

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import json
        import numpy as np
        from clinica.utils.stream import cprint

        # Reading BIDS files
        # ==================
        read_node = npe.Node(name="ReadingBIDS",
                             interface=nutil.IdentityInterface(
                                     fields=self.get_input_fields(),
                                     mandatory_inputs=True))
        # I remove the 'sub-' prefix that is not considered by the pybids'
        # layout object.
        subject_regex = '|'.join(s[4:] for s in self.subjects)

        if ('unwarping' in self.parameters) and self.parameters['unwarping']:
            read_node.inputs.magnitude1 = self.bids_layout.get(
                    return_type='file',
                    type='magnitude1',
                    extensions='nii.gz',
                    subject=subject_regex)
            read_node.inputs.phasediff = self.bids_layout.get(
                    return_type='file',
                    type='phasediff',
                    extensions='nii.gz',
                    subject=subject_regex)
        read_node.inputs.bold = self.bids_layout.get(return_type='file',
                                                     type='bold',
                                                     extensions='nii.gz',
                                                     subject=subject_regex)
        read_node.inputs.T1w = self.bids_layout.get(return_type='file',
                                                    type='T1w',
                                                    extensions='nii.gz',
                                                    subject=subject_regex)

        # Reading BIDS json
        # =================

        read_node.inputs.et = []
        read_node.inputs.blipdir = []
        read_node.inputs.tert = []
        read_node.inputs.time_repetition = []
        read_node.inputs.num_slices = []
        read_node.inputs.slice_order = []
        read_node.inputs.ref_slice = []
        read_node.inputs.time_acquisition = []

        for i in range(len(self.subjects)):
            cprint('Loading subject "{sub}"...'.format(sub=self.subjects[i]))

            if self.parameters['unwarping']:
                # From phasediff json file
                phasediff_json = self.bids_layout.get(return_type='file',
                                                      type='phasediff',
                                                      extensions='json',
                                                      subject=self.subjects[i][
                                                              4:])
                with open(phasediff_json[0]) as json_file:
                    data = json.load(json_file)
                    # SPM echo times
                    read_node.inputs.et.append([data['EchoTime1'],
                                                data['EchoTime2']])
                    # SPM blip direction
                    # TODO: Verifiy that it is the correct way to get the
                    # blipdir
                    blipdir_raw = data['PhaseEncodingDirection']
                    if len(blipdir_raw) > 1 and blipdir_raw[1] == '-':
                        read_node.inputs.blipdir.append(-1)
                    else:
                        read_node.inputs.blipdir.append(1)

            # From func json file
            func_json = self.bids_layout.get(return_type='file',
                                             type='bold',
                                             extensions='json',
                                             subject=self.subjects[i][4:])
            with open(func_json[0]) as json_file:
                data = json.load(json_file)
                # SPM Total readout time
                read_node.inputs.tert.append(
                        1 / data['BandwidthPerPixelPhaseEncode'])
                # SPM Repetition time
                read_node.inputs.time_repetition.append(data['RepetitionTime'])
                # Number of slices
                slice_timing = data['SliceTiming']
                read_node.inputs.num_slices.append(len(slice_timing))
                # Slice order
                slice_order = np.argsort(slice_timing) + 1
                read_node.inputs.slice_order.append(slice_order.tolist())
                read_node.inputs.ref_slice.append(np.argmin(slice_timing) + 1)
                read_node.inputs.time_acquisition.append(
                        data['RepetitionTime'] - data['RepetitionTime']
                        / float(len(slice_timing)))

            cprint(read_node.inputs)

        if ('unwarping' in self.parameters) and self.parameters['unwarping']:
            self.connect([
                # Reading BIDS json
                (read_node, self.input_node, [('et', 'et')]),
                (read_node, self.input_node, [('blipdir', 'blipdir')]),
                (read_node, self.input_node, [('tert', 'tert')]),
                # Reading BIDS files
                (read_node, self.input_node, [('phasediff', 'phasediff')]),
                (read_node, self.input_node, [('magnitude1', 'magnitude1')]),
            ])

        self.connect([
            # Reading BIDS json
            (read_node, self.input_node,
             [('time_repetition', 'time_repetition')]),
            (read_node, self.input_node, [('num_slices', 'num_slices')]),
            (read_node, self.input_node, [('slice_order', 'slice_order')]),
            (read_node, self.input_node, [('ref_slice', 'ref_slice')]),
            (read_node, self.input_node,
             [('time_acquisition', 'time_acquisition')]),
            # Reading BIDS files
            (read_node, self.input_node, [('bold', 'bold')]),
            (read_node, self.input_node, [('T1w', 'T1w')]),
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """

        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio

        # Writing CAPS
        # ============
        write_node = npe.MapNode(name='WritingCAPS',
                                 iterfield=['container'] + self.get_output_fields(),
                                 interface=nio.DataSink(
                                         infields=self.get_output_fields()))
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.parameterization = False
        write_node.inputs.container = [
            'subjects/' + self.subjects[i] + '/' + self.sessions[i] +
            '/fmri/preprocessing' for i in range(len(self.subjects))]
        write_node.inputs.remove_dest_dir = True
        if ('freesurfer_brain_mask' in self.parameters) and \
                not (self.parameters['freesurfer_brain_mask']):
            write_node.inputs.regexp_substitutions = [
                (r't1_brain_mask/c3(.+)_maths_dil_ero_thresh_fillh\.nii\.gz$',
                 r'\1_brainmask.nii.gz'),
                (r'mc_params/rp_a(.+)\.txt$', r'\1_motion.tsv'),
                (r'native_fmri/[u|r]a(.+)\.nii.gz$',
                 r'\1_space-meanBOLD_preproc.nii.gz'),
                (r't1_fmri/r[u|r]a(.+)\.nii.gz$',
                 r'\1_space-T1w_preproc.nii.gz'),
                (r'mni_fmri/wr[u|r]a(.+)\.nii.gz$',
                 r'\1_space-Ixi549Space_preproc.nii.gz'),
                (r'mni_smoothed_fmri/swr[u|r]a(.+)\.nii.gz$',
                 r'\1_space-Ixi549Space_fwhm-' + 'x'.join(
                     map(str, self.parameters[
                         'full_width_at_half_maximum'])) + '_preproc.nii.gz'),
                # I don't know why it's adding this empty folder, so I remove
                # it:
                (r'trait_added', r''),
            ]
        else:
            write_node.inputs.regexp_substitutions = [
                (r't1_brain_mask/(.+)\.nii\.gz$', r'\1_brainmask.nii.gz'),
                (r'mc_params/rp_a(.+)\.txt$', r'\1_motion.tsv'),
                (r'native_fmri/[u|r]a(.+)\.nii.gz$',
                 r'\1_space-meanBOLD_preproc.nii.gz'),
                (r't1_fmri/r[u|r]a(.+)\.nii.gz$',
                 r'\1_space-T1w_preproc.nii.gz'),
                (r'mni_fmri/wr[u|r]a(.+)\.nii.gz$',
                 r'\1_space-Ixi549Space_preproc.nii.gz'),
                (r'mni_smoothed_fmri/swr[u|r]a(.+)\.nii.gz$',
                 r'\1_space-Ixi549Space_fwhm-' + 'x'.join(
                     map(str, self.parameters[
                         'full_width_at_half_maximum'])) +
                 '_preproc.nii.gz'),
                # I don't know why it's adding this empty folder, so I remove
                # it:
                (r'trait_added', r''),
            ]

        # fMRI images in the subject's T1 native space are large, we add it
        # only if specified:
        if ('t1_native_space' in self.parameters) and self.parameters[
                't1_native_space']:
            self.connect([
                (self.output_node, write_node, [('t1_fmri', 't1_fmri')]),
            ])

        self.connect([
            # Writing CAPS
            (self.output_node, write_node,
             [('t1_brain_mask', 't1_brain_mask')]),
            (self.output_node, write_node,
             [('mc_params', 'mc_params')]),
            (self.output_node, write_node,
             [('native_fmri', 'native_fmri')]),
            (self.output_node, write_node,
             [('mni_fmri', 'mni_fmri')]),
            (self.output_node, write_node,
             [('mni_smoothed_fmri', 'mni_smoothed_fmri')]),
        ])

    def check_custom_dependencies(self):
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """

        import clinica.pipelines.fmri_preprocessing.fmri_preprocessing_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.interfaces.spm as spm
        import nipype.pipeline.engine as npe
        from clinica.utils.io import zip_nii, unzip_nii

        # Zipping
        # =======
        unzip_node = npe.MapNode(name='Unzipping',
                                 iterfield=['in_file'],
                                 interface=nutil.Function(
                                         input_names=['in_file'],
                                         output_names=[
                                             'out_file'],
                                         function=unzip_nii))

        unzip_T1w = unzip_node.clone('UnzippingT1w')
        unzip_phasediff = unzip_node.clone('UnzippingPhasediff')
        unzip_bold = unzip_node.clone('UnzippingBold')
        unzip_magnitude1 = unzip_node.clone('UnzippingMagnitude1')

        # FieldMap calculation
        # ====================
        if self.parameters['unwarping']:
            fm_node = npe.MapNode(name="FieldMapCalculation",
                                  iterfield=['phase', 'magnitude', 'epi',
                                             'et', 'blipdir', 'tert'],
                                  interface=spm.FieldMap())

        # Slice timing correction
        # =======================
        st_node = npe.MapNode(name="SliceTimingCorrection",
                              iterfield=['in_files', 'time_repetition',
                                         'slice_order', 'num_slices',
                                         'ref_slice', 'time_acquisition'],
                              interface=spm.SliceTiming())

        # Motion correction and unwarping
        # ===============================

        if self.parameters['unwarping']:
            mc_node = npe.MapNode(name="MotionCorrectionUnwarping",
                                  iterfield=["scans", "pmscan"],
                                  interface=spm.RealignUnwarp())
            mc_node.inputs.register_to_mean = True
            mc_node.inputs.reslice_mask = False
        else:
            mc_node = npe.MapNode(name="MotionCorrection",
                                  iterfield=["in_files"],
                                  interface=spm.Realign())
            mc_node.inputs.register_to_mean = True

        # Brain extraction
        # ================
        import os.path as path
        from nipype.interfaces.freesurfer import MRIConvert
        if self.parameters['freesurfer_brain_mask']:
            brain_masks = [path.join(self.caps_directory, 'subjects',
                                     self.subjects[i], self.sessions[i],
                                     't1/freesurfer_cross_sectional',
                                     self.subjects[i] + '_' + self.sessions[i],
                                     'mri/brain.mgz')
                           for i in range(len(self.subjects))]
            conv_brain_masks = [str(self.subjects[i] + '_' + self.sessions[i] +
                                    '.nii')
                                for i in range(len(self.subjects))]
            bet_node = npe.MapNode(interface=MRIConvert(),
                                   iterfield=["in_file", "out_file"],
                                   name="BrainConversion")
            bet_node.inputs.in_file = brain_masks
            bet_node.inputs.out_file = conv_brain_masks
            bet_node.inputs.out_type = 'nii'
        else:
            bet_node = utils.BrainExtractionWorkflow(name="BrainExtraction")

        # Registration
        # ============
        reg_node = npe.MapNode(interface=spm.Coregister(),
                               iterfield=["apply_to_files", "source", "target"],
                               name="Registration")

        # Normalization
        # =============
        norm_node = npe.MapNode(interface=spm.Normalize12(),
                                iterfield=['image_to_align', 'apply_to_files'],
                                name='Normalization')

        # Smoothing
        # =========
        smooth_node = npe.MapNode(interface=spm.Smooth(),
                                  iterfield=['in_files'],
                                  name='Smoothing')
        smooth_node.inputs.fwhm = self.parameters['full_width_at_half_maximum']

        # Zipping
        # =======
        zip_node = npe.MapNode(name='Zipping',
                               iterfield=['in_file'],
                               interface=nutil.Function(input_names=['in_file'],
                                                        output_names=[
                                                            'out_file'],
                                                        function=zip_nii))

        zip_bet_node = zip_node.clone('ZippingBET')
        zip_mc_node = zip_node.clone('ZippingMC')
        zip_reg_node = zip_node.clone('ZippingRegistration')
        zip_norm_node = zip_node.clone('ZippingNormalization')
        zip_smooth_node = zip_node.clone('ZippingSmoothing')

        # Connections
        # ===========

        if self.parameters['freesurfer_brain_mask']:
            self.connect([
                # Brain extraction
                (bet_node, reg_node, [('out_file', 'target')]),
                (bet_node, zip_bet_node, [('out_file', 'in_file')]),
            ])
        else:
            self.connect([
                # Brain extraction
                (unzip_T1w, bet_node, [('out_file', 'Segmentation.data')]),
                (unzip_T1w, bet_node, [('out_file', 'ApplyMask.in_file')]),
                (bet_node, reg_node, [('ApplyMask.out_file', 'target')]),
                (bet_node, zip_bet_node, [('Fill.out_file', 'in_file')]),
            ])

        if self.parameters['unwarping']:
            self.connect([
                # FieldMap calculation
                (self.input_node, fm_node, [('et', 'et')]),
                (self.input_node, fm_node, [('blipdir', 'blipdir')]),
                (self.input_node, fm_node, [('tert', 'tert')]),
                (self.input_node, unzip_phasediff, [('phasediff', 'in_file')]),
                (self.input_node, unzip_magnitude1,
                 [('magnitude1', 'in_file')]),
                (unzip_magnitude1, fm_node, [('out_file', 'magnitude')]),
                (unzip_phasediff, fm_node, [('out_file', 'phase')]),
                (unzip_bold, fm_node, [('out_file', 'epi')]),
                # Motion correction and unwarping
                (st_node, mc_node, [('timecorrected_files', 'scans')]),
                (fm_node, mc_node, [('vdm', 'pmscan')]),
                (mc_node, reg_node, [('realigned_unwarped_files', 'apply_to_files')]),
                (mc_node, zip_mc_node, [('realigned_unwarped_files', 'in_file')]),
            ])
        else:
            self.connect([
                # Motion correction and unwarping
                (st_node, mc_node, [('timecorrected_files', 'in_files')]),
                (mc_node, reg_node, [('realigned_files', 'apply_to_files')]),
                (mc_node, zip_mc_node, [('realigned_files', 'in_file')]),
            ])
        self.connect([
            # Unzipping
            (self.input_node, unzip_T1w, [('T1w', 'in_file')]),
            (self.input_node, unzip_bold, [('bold', 'in_file')]),
            # Slice timing correction
            (unzip_bold, st_node, [('out_file', 'in_files')]),
            (self.input_node, st_node,
             [('time_repetition', 'time_repetition')]),
            (self.input_node, st_node, [('num_slices', 'num_slices')]),
            (self.input_node, st_node, [('slice_order', 'slice_order')]),
            (self.input_node, st_node, [('ref_slice', 'ref_slice')]),
            (self.input_node, st_node,
             [('time_acquisition', 'time_acquisition')]),
            # Registration
            (mc_node, reg_node, [('mean_image', 'source')]),
            # Normalization
            (unzip_T1w, norm_node, [('out_file', 'image_to_align')]),
            (reg_node, norm_node, [('coregistered_files', 'apply_to_files')]),
            # Smoothing
            (norm_node, smooth_node, [('normalized_files', 'in_files')]),
            # Zipping
            (reg_node, zip_reg_node, [('coregistered_files', 'in_file')]),
            (norm_node, zip_norm_node, [('normalized_files', 'in_file')]),
            (smooth_node, zip_smooth_node, [('smoothed_files', 'in_file')]),
            # Returning output
            (zip_bet_node, self.output_node, [('out_file', 't1_brain_mask')]),
            (mc_node, self.output_node,
             [('realignment_parameters', 'mc_params')]),
            (zip_mc_node, self.output_node, [('out_file', 'native_fmri')]),
            (zip_reg_node, self.output_node, [('out_file', 't1_fmri')]),
            (zip_norm_node, self.output_node, [('out_file', 'mni_fmri')]),
            (zip_smooth_node, self.output_node, [('out_file',
                                                  'mni_smoothed_fmri')]),
        ])



# coding: utf8

import clinica.engine as ce


class fMRIPreprocessingCLI(ce.CmdParser):
    """
    """

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 'fmri-preprocessing'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Preprocessing of raw fMRI datasets:\nhttp://clinica.run/doc/Pipelines/fMRI_Preprocessing/'

    def define_options(self):
        """Define the sub-command arguments
        """

        from clinica.engine.cmdparser import PIPELINE_CATEGORIES

        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')

        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-tsv", "--subjects_sessions_tsv",
                                 help='TSV file containing a list of subjects with their sessions.')
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results.')
        clinica_opt.add_argument("-np", "--n_procs", type=int,
                                 help='Number of processors to run in parallel.')
#        clinica_opt.add_argument("-sl", "--slurm", action='store_true',
#                                 help='Run the pipelines using SLURM.')
#        clinica_opt.add_argument("-sa", "--sbatch_args",
#                                 help='SLURM\'s sbatch tool arguments.')

        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-fwhm", "--full_width_at_half_maximum",
                              metavar="FWHM",
                              nargs=3, type=int, default=[8, 8, 8],
                              help="A list of integers specifying the full width at half maximum (FWHM) filter size in the three directions x y z, in millimeters, used to smooth the images (default: --full_width_at_half_maximum 8 8 8).")
        optional.add_argument("-t1s", "--t1_native_space", action='store_true',
                              help="When specified, the corrected fMRI data registered to the native T1 space of each subject are saved. This might be useful when dealing with parcellations that are in the native T1 space.")
        optional.add_argument("-fsbm", "--freesurfer_brain_mask",
                              action='store_true',
                              help="When specified, the pipeline uses the brain extraction generated by FreeSurfer during the t1-freesurfer pipeline. Otherwise, it computes it using the SPM Registration tool followed by empirically-estimated operations performed by FSL.")
        optional.add_argument("-u", "--unwarping",
                              action='store_true',
                              help="When specified, the unwarping correction step is executed. This will add SPM's Unwarping to the Realign step (magnitude and phasediff files are necessary in the BIDS directory).")

    def run_command(self, args):
        """
        """
        from clinica.utils.stream import cprint
        from tempfile import mkdtemp
        from .fmri_preprocessing_pipeline import fMRIPreprocessing

        pipeline = fMRIPreprocessing(bids_directory=self.absolute_path(args.bids_directory),
                                     caps_directory=self.absolute_path(args.caps_directory),
                                     tsv_file=self.absolute_path(args.subjects_sessions_tsv))
        pipeline.parameters = {
            'full_width_at_half_maximum' : args.full_width_at_half_maximum,
            't1_native_space'            : args.t1_native_space,
            'freesurfer_brain_mask'      : args.freesurfer_brain_mask,
            'unwarping'                  : args.unwarping
        }
        if args.working_directory is None:
            args.working_directory = mkdtemp()
        pipeline.base_dir = self.absolute_path(args.working_directory)
        if args.n_procs:
            pipeline.run(plugin='MultiProc', plugin_args={'n_procs': args.n_procs})
#        elif args.slurm:
#            pipeline.run(plugin='SLURMGraph', plugin_args={
#                'dont_resubmit_completed_jobs': True, 'sbatch_args':
#                    args.sbatch_args})
        else:
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8

import clinica.engine as ce


class T1VolumeParcellationCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline.
        """
        self._name = 't1-volume-parcellation'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Computation of mean GM concentration for a set of regions:\n' \
                            + 'http://clinica.run/doc/Pipelines/T1_Volume/'

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-tsv", "--subjects_sessions_tsv",
                                 help='TSV file containing a list of subjects with their sessions.')
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results')
        clinica_opt.add_argument("-np", "--n_procs",
                                 type=int,
                                 help='Number of cores used to run in parallel')
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-m", "--modulation",
                              metavar="[on|off]", default='on',
                              help='Specify if modulation must be enabled (default: --modulation on')
        list_atlases = ['AAL2', 'LPBA40', 'Neuromorphometrics', 'AICHA', 'Hammers']
        advanced.add_argument("-atlases", "--atlases",
                              nargs='+', type=str, metavar='',
                              default=list_atlases, choices=list_atlases,
                              help='A list of atlases used to calculate the regional mean GM concentrations (default: all atlases i.e. --atlases AAL2 AICHA Hammers LPBA40 Neuromorphometrics).')

    def run_command(self, args):
        """
        """
        from tempfile import mkdtemp
        from clinica.utils.stream import cprint
        from clinica.pipelines.t1_volume_parcellation.t1_volume_parcellation_pipeline import T1VolumeParcellation

        pipeline = T1VolumeParcellation(
             caps_directory=self.absolute_path(args.caps_directory),
             tsv_file=self.absolute_path(args.subjects_sessions_tsv))
        assert args.modulation in ['on', 'off']
        pipeline.parameters = {
            'group_id': args.group_id,
            'atlases': args.atlases,
            'wd': self.absolute_path(args.working_directory),
            'n_procs': args.n_procs,
            'modulate': args.modulation
        }

        if args.working_directory is None:
            args.working_directory = mkdtemp()
        pipeline.base_dir = self.absolute_path(args.working_directory)

        if args.n_procs:
            pipeline.run(plugin='MultiProc', plugin_args={'n_procs': args.n_procs})
        else:
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8

"""
Convert the AIBL dataset (http://www.aibl.csiro.au/) into BIDS.
"""

__author__ = "Simona Bottani"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Simona Bottani"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Simona Bottani"
__email__ = "simona.bottani@icm-institute.org"
__status__ = "Development"


def convert_images(path_to_dataset, path_to_csv, bids_dir):

    # Conversion of the entire dataset in BIDS
    from clinica.utils.stream import cprint
    from os.path import exists
    from colorama import Fore

    from clinica.iotools.converters.aibl_to_bids.aibl_utils import paths_to_bids
    list_of_created_files = []

    for modality in ['t1', 'av45', 'flute', 'pib']:
        list_of_created_files.append(paths_to_bids(path_to_dataset,
                                                   path_to_csv,
                                                   bids_dir,
                                                   modality))

    error_string = ''
    for modality_list in list_of_created_files:
        for file in modality_list:
            if not exists(str(file)):
                error_string = error_string + str(file) + '\n'
    if error_string != '':
        cprint(Fore.RED + 'The following file were not converted '
               + ' (nan means no path was found):\n'
               + error_string
               + Fore.RESET)


def convert_clinical_data(bids_dir, path_to_csv):
    from os.path import exists
    # clinical specifications in BIDS
    from os.path import join, split, realpath
    from clinica.iotools.converters.aibl_to_bids.aibl_utils import create_participants_df_AIBL, \
        create_sessions_dict_AIBL

    clinical_spec_path = join(split(realpath(__file__))[0], '../../data/clinical_specifications.xlsx')
    if not exists(clinical_spec_path):
        raise FileNotFoundError(clinical_spec_path + ' file not found ! This is an internal file of Clinica.')

    create_participants_df_AIBL(bids_dir, clinical_spec_path, path_to_csv, delete_non_bids_info=True)
    create_sessions_dict_AIBL(bids_dir, path_to_csv, clinical_spec_path)



# coding: utf8

__author__ = "Junhao Wen"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Junhao Wen", "Alexandre Routier"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Junhao Wen"
__email__ = "Junhao.Wen@inria.fr"
__status__ = "Development"

import clinica.pipelines.engine as cpe


class T1FreeSurferCrossSectional(cpe.Pipeline):
    """Performs cortical surface extraction, segmentation of subcortical
    structures, cortical thickness estimation, spatial normalization onto the
    FreeSurfer surface template (FsAverage), and parcellation of cortical
    regions based on different atlases using the FreeSurfer recon-all command.
    It also generate TSV files containing a summary of the regional statistics
    (e.g. regional volume, mean cortical thickness) to ease subsequent
    statistical analysis.

    Args:
        bids_directory: Path to the BIDS directory.
        caps_directory: Path to the CAPS directory.
        tsv: TSV file containing the subjects with their sessions
        ras: additional flags for recon-all command line, default will be -qcache
        wd: Temporary directory to store pipelines intermediate results
        np: Number of cores used to run in parallel

    Returns:
        A clinica pipeline object containing the T1FreeSurferCrossSectional pipeline.
    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """
        return ['recon_all_args', 'subject_list', 'session_list', 'subject_id', 'subject_dir', 'anat_t1']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """
        return ['subject_id']

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """
        import clinica.pipelines.t1_freesurfer_cross_sectional.t1_freesurfer_cross_sectional_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.stream import cprint

        # This node is supposedly used to load BIDS inputs when this pipelines is
        # not already connected to the output of a previous Clinica pipelines.
        # For the purpose of the example, we simply read input arguments given
        # by the command line interface and transmitted here through the
        # `self.parameters` dictionary and pass it to the `self.input_node` to
        # further by used as input of the core nodes.

        # An IdentityInterface to get the input fields for the pipelines
        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True))
        read_parameters_node.inputs.recon_all_args = self.parameters['recon_all_args']

        # check if a specific subject has already run recon-all, and return the related lists in the tsv file
        dataprepr = npe.Node(name='dataprepr',
                             interface=nutil.Function(
                                 input_names=['output_dir', 'subject_list', 'session_list'],
                                 output_names=['nouse0', 'nouse1', 'subject_dir', 'subject_id',
                                               'subject_list', 'session_list'],
                                 function=utils.get_dirs_check_reconalled))
        dataprepr.inputs.output_dir = self.caps_directory
        dataprepr.inputs.subject_list = self.subjects
        dataprepr.inputs.session_list = self.sessions

        # use pybids to grab the anat_t1 images based on the tsv file
        datagrabbernode = npe.Node(name='datagrabbernode',
                                   interface=nutil.Function(
                                       function=utils.bids_datagrabber,
                                       input_names=['input_dir', 'subject_list', 'session_list'],
                                       output_names=['anat_t1']))
        datagrabbernode.inputs.input_dir = self.bids_directory

        self.connect([
            (read_parameters_node, self.input_node, [('recon_all_args', 'recon_all_args')]),
            (dataprepr, datagrabbernode, [('subject_list', 'subject_list')]),
            (dataprepr, datagrabbernode, [('session_list', 'session_list')]),
            (dataprepr, self.input_node, [('subject_list', 'subject_list')]),
            (dataprepr, self.input_node, [('session_list', 'session_list')]),
            (dataprepr, self.input_node, [('subject_id', 'subject_id')]),
            (dataprepr, self.input_node, [('subject_dir', 'subject_dir')]),
            (datagrabbernode, self.input_node, [('anat_t1', 'anat_t1')]),
        ])

        cprint('The pipeline will last approximately 10 hours per image.')

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """
        import clinica.pipelines.t1_freesurfer_cross_sectional.t1_freesurfer_cross_sectional_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.freesurfer.preprocess import ReconAll

        # MapNode to check out if we need -cw256 for every subject, and -qcache is default for every subject.
        flagnode = npe.MapNode(name='flagnode',
                               iterfield=['t1_list'],
                               interface=nutil.Function(
                                   input_names=['t1_list', 'recon_all_args'],
                                   output_names=['output_flags'],
                                   function=utils.check_fov))

        # MapNode to transfer every subject's flag to string.
        create_flags = npe.MapNode(interface=nutil.Function(
            input_names=['input_flags'],
            output_names=['output_str'],
            function=utils.create_flags_str),
            name='create_flags_string',
            iterfield=['input_flags'])

        # MapNode to implement recon-all.
        recon_all = npe.MapNode(interface=ReconAll(),
                                name='recon_all',
                                iterfield=['subject_id', 'T1_files', 'subjects_dir', 'flags'])
        recon_all.inputs.directive = 'all'

        # MapNode to create the statistical tsv file for each subject
        tsvmapnode = npe.MapNode(name='tsvmapnode',
                                 iterfield=['subject_id'],
                                 interface=nutil.Function(
                                     input_names=['subject_id', 'output_dir'],
                                     output_names=[],
                                     function=utils.write_tsv_files,
                                     imports=['import os', 'import errno']))
        tsvmapnode.inputs.output_dir = self.caps_directory

        print_begin_message = npe.MapNode(
            interface=nutil.Function(
                input_names=['subject_id'],
                function=utils.print_begin_pipeline),
            iterfield='subject_id',
            name='Write-Begin_Message')

        print_end_message = npe.MapNode(
            interface=nutil.Function(
                input_names=['subject_id', 'final_file'],
                function=utils.print_end_pipeline),
            iterfield='subject_id',
            name='Write-End_Message')

        # Connection
        # ==========
        self.connect([
            (self.input_node, print_begin_message, [('subject_id', 'subject_id')]),
            (self.input_node, flagnode, [('recon_all_args', 'recon_all_args')]),
            (self.input_node, flagnode, [('anat_t1',        't1_list')]),
            (flagnode, create_flags, [('output_flags', 'input_flags')]),
            (create_flags,    recon_all, [('output_str',  'flags')]),
            (self.input_node, recon_all, [('subject_dir', 'subjects_dir')]),
            (self.input_node, recon_all, [('subject_id',  'subject_id')]),
            (self.input_node, recon_all, [('anat_t1',     'T1_files')]),
            #            (self.input_node, lognode, [('subject_list', 'subject_list')]),
            #            (self.input_node, lognode, [('session_list', 'session_list')]),
            (recon_all, tsvmapnode, [('subject_id', 'subject_id')]),
            #            (recon_all, lognode, [('subject_id', 'subject_id')]),
            (recon_all, self.output_node, [('subject_id', 'subject_id')]),
            (self.input_node, print_end_message, [('subject_id', 'subject_id')]),
            (recon_all,       print_end_message, [('subject_id', 'final_file')]),
        ])

        return self



# coding: utf-8

"""
 Module for converting DWI of ADNI
"""

__author__ = "Jorge Samper-Gonzalez and Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_dwi(source_dir, csv_dir, dest_dir, subjs_list=None):
    """

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination directory
        subjs_list: subjects list

    Returns:

    """
    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of DWI images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_dwi_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of DWI images found. Exporting images into BIDS ...')
    dwi_paths_to_bids(images, dest_dir)
    cprint(Fore.GREEN + 'DWI conversion done.' + Fore.RESET)


def compute_dwi_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute paths to DWI images to convert to BIDS

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination directory
        subjs_list: subjects list

    Returns:
        images: pandas dataframe that contains the path to all the dwi images to convert

    """
    import pandas as pd
    import operator
    from os import path, walk, mkdir
    from functools import reduce

    dwi_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                  'Study_ID', 'Series_ID', 'Image_ID', 'Field_Strength', 'Scanner', 'Enhanced']

    dwi_df = pd.DataFrame(columns=dwi_col_df)

    adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')

    if path.exists(path.join(csv_dir, 'IDA_MR_Metadata_Listing.csv')):
        new_download = False
    else:
        new_download = True

    # If IDA_MR_Metadata_Listing.csv has been added manually, that will cause a
    # bug
    new_download = True

    if new_download:
        ida_meta_path = path.join(csv_dir, 'MRILIST.csv')
    else:
        ida_meta_path = path.join(csv_dir, 'IDA_MR_Metadata_Listing.csv')
    mri_qc_path = path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv')

    adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
    ida_meta = pd.io.parsers.read_csv(ida_meta_path, sep=',')

    if new_download:
        ida_meta = ida_meta[ida_meta.SEQUENCE.map(lambda x: x.lower().find('dti') > -1)]
    else:
        ida_meta = ida_meta[ida_meta.Sequence.map(lambda x: x.lower().find('dti') > -1)]

    mri_qc = pd.io.parsers.read_csv(mri_qc_path, sep=',')
    mri_qc = mri_qc[mri_qc.series_type == 'DTI']

    for subj in subjs_list:
        # print 'Computing path for subj', subj

        adnimerge_subj = adni_merge[adni_merge.PTID == subj]

        # Sort the values by examination date
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')
        if new_download:
            ida_meta_subj = ida_meta[ida_meta.SUBJECT == subj]
            ida_meta_subj = ida_meta_subj.sort_values('SCANDATE')
        else:
            ida_meta_subj = ida_meta[ida_meta.Subject == subj]
            ida_meta_subj = ida_meta_subj.sort_values('Scan Date')
        mri_qc_subj = mri_qc[mri_qc.RID == int(subj[-4:])]

        if new_download:
            visits = visits_to_timepoints_dwi_refactoring(subj, ida_meta_subj, adnimerge_subj)
        else:
            visits = visits_to_timepoints_dwi(subj, ida_meta_subj, adnimerge_subj)

        keys = visits.keys()
        # What is the purpose of the following line ?
        # keys.sort()

        for visit_info in visits.keys():
            visit_str = visits[visit_info]

            if new_download:
                visit_ida_meta = ida_meta_subj[ida_meta_subj.VISIT == visit_str]
                axial_ida_meta = visit_ida_meta[visit_ida_meta.SEQUENCE.map(lambda x: x.lower().find('enhanced') < 0)]
                enhanced_ida_meta = visit_ida_meta[visit_ida_meta.SEQUENCE.map(lambda x: x.lower().find('enhanced') > -1)]
                axial = dwi_image_refactoring(subj, visit_info[0], visits[visit_info], axial_ida_meta, mri_qc_subj, False)
                enhanced = dwi_image_refactoring(subj, visit_info[0], visits[visit_info], enhanced_ida_meta, mri_qc_subj, True)
            else:
                visit_ida_meta = ida_meta_subj[ida_meta_subj.Visit == visit_str]
                axial_ida_meta = visit_ida_meta[visit_ida_meta.Sequence.map(lambda x: x.lower().find('enhanced') < 0)]
                enhanced_ida_meta = visit_ida_meta[visit_ida_meta.Sequence.map(lambda x: x.lower().find('enhanced') > -1)]
                axial = dwi_image(subj, visit_info[0], visits[visit_info], axial_ida_meta, mri_qc_subj, False)
                enhanced = dwi_image(subj, visit_info[0], visits[visit_info], enhanced_ida_meta, mri_qc_subj, True)

            if axial is not None:
                row_to_append = pd.DataFrame(axial, index=['i', ])
                dwi_df = dwi_df.append(row_to_append, ignore_index=True)

            if enhanced is not None:
                row_to_append = pd.DataFrame(enhanced, index=['i', ])
                dwi_df = dwi_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [('029_S_2395', 'm60'),
                         ('029_S_0824', 'm108'),
                         ('029_S_0914', 'm108'),
                         ('027_S_2219', 'm36'),
                         ('129_S_2332', 'm12'),
                         ('029_S_4384', 'm48'),
                         ('029_S_4385', 'm48'),
                         ('029_S_4585', 'm48'),
                         ('016_S_4591', 'm24'),
                         ('094_S_4630', 'm06'),
                         ('094_S_4649', 'm06'),
                         ('029_S_5219', 'm24'),
                         ('094_S_2238', 'm48'),
                         ('129_S_4287', 'bl'),
                         ('007_S_4611', 'm03'),
                         ('016_S_4638', 'bl'),
                         ('027_S_5118', 'bl'),
                         ('098_S_4018', 'bl'),
                         ('098_S_4003', 'm12'),
                         ('016_S_4584', 'm24'),
                         ('016_S_5007', 'm12'),
                         ('129_S_2347', 'm06'),
                         ('129_S_4220', 'bl'),
                         ('007_S_2058', 'm12'),
                         ('016_S_2007', 'm06')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((dwi_df.Enhanced is False)
                             & (dwi_df.Subject_ID == conv_error[0])
                             & (dwi_df.VISCODE == conv_error[1]))

    indices_to_remove = dwi_df.index[reduce(operator.or_, error_indices, False)]
    dwi_df.drop(indices_to_remove, inplace=True)

    images = dwi_df
    is_dicom = []
    nifti_paths = []
    count = 0

    for row in images.iterrows():
        image = row[1]
        seq_path = path.join(source_dir, str(image.Subject_ID), image.Sequence)
        count += 1
        series_path = ''
        s = 'S' + str(image.Series_ID)

        for (dirpath, dirnames, filenames) in walk(seq_path):
            found = False
            for d in dirnames:
                if d == s:
                    series_path = path.join(dirpath, d)
                    found = True
                    break
            if found:
                break

        nifti_path = series_path
        dicom = True

        is_dicom.append(dicom)
        nifti_paths.append(nifti_path)

    images.loc[:, 'Is_Dicom'] = pd.Series(is_dicom, index=images.index)
    images.loc[:, 'Path'] = pd.Series(nifti_paths, index=images.index)

    dwi_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(dwi_tsv_path):
        mkdir(dwi_tsv_path)
    images.to_csv(path.join(dwi_tsv_path, 'dwi_paths.tsv'), sep='\t', index=False)

    return images


def dwi_paths_to_bids(images, dest_dir, mod_to_update=False):
    """
    Convert DWI images

    Args:
        images: dataframe returned by the method compute_dwi_paths
        dest_dir: path to the destination directory
        mod_to_update: if is true and an image is already existing it will overwrite the old version

    """
    from multiprocessing import cpu_count, Pool, Value
    from functools import partial

    counter = None

    def init(args):
        ''' store the counter for later use '''
        global counter
        counter = args

    subjs_list = images['Subject_ID'].unique()
    counter = Value('i', 0)
    total = len(images)
    partial_generate_subject_files = partial(generate_subject_files,
                                             images=images,
                                             dest_dir=dest_dir,
                                             mod_to_update=mod_to_update)
    poolrunner = Pool(cpu_count(), initializer=init, initargs=(counter,))
    poolrunner.map(partial_generate_subject_files, subjs_list)
    del counter


def generate_subject_files(subj, images, dest_dir, mod_to_update):
    import clinica.iotools.bids_utils as bids
    import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils
    from clinica.utils.stream import cprint
    import subprocess
    import os
    import shutil
    from os import path
    from glob import glob

    global counter
    alpha_id = bids.remove_space_and_symbols(subj)
    bids_id = 'sub-ADNI' + alpha_id
    # Extract the list of sessions available from the dwi paths files, removing the duplicates
    sess_list = images[(images['Subject_ID'] == subj)]['VISCODE'].unique()

    if not os.path.exists(path.join(dest_dir, bids_id)):
        os.mkdir(path.join(dest_dir, bids_id))

    # For each session available, create the folder if doesn't exist and convert the files
    for ses in sess_list:
        with counter.get_lock():
            counter.value += 1
        cprint('[DWI] Processing subject ' + str(subj)
               + ' - session ' + ses + ', ' + str(counter.value)
               + ' / ' + str(len(images)))
        ses_bids = adni_utils.viscode_to_session(ses)
        bids_ses_id = 'ses-' + ses_bids
        bids_file_name = bids_id + '_ses-' + ses_bids
        ses_path = path.join(dest_dir, bids_id, bids_ses_id)

        if mod_to_update:
            if os.path.exists(path.join(ses_path, 'dwi')):
                shutil.rmtree(path.join(ses_path, 'dwi'))

        if not os.path.exists(ses_path):
            os.mkdir(ses_path)

        dwi_info = images[
            (images['Subject_ID'] == subj) & (images['VISCODE'] == ses)]

        # For the same subject, same session there could be multiple dwi with different acq label
        for j in range(len(dwi_info)):
            dwi_subj = dwi_info.iloc[j]
            # TODO For now in CLINICA we ignore Enhanced DWI.
            if dwi_subj['Enhanced']:
                continue
            if type(dwi_subj['Path']) != float and dwi_subj['Path'] != '':
                if not os.path.exists(path.join(ses_path, 'dwi')):
                    os.mkdir(path.join(ses_path, 'dwi'))
                dwi_path = dwi_subj['Path']

                bids_name = bids_file_name + '_acq-' + ('axialEnhanced'
                                                        if dwi_subj['Enhanced'] else 'axial') + '_dwi'

                bids_dest_dir = path.join(ses_path, 'dwi')

                if not os.path.exists(bids_dest_dir):
                    os.mkdir(dest_dir)
                command = 'dcm2niix -b n -z y -o ' + bids_dest_dir + ' -f ' + bids_name + ' ' + dwi_path
                subprocess.run(command,
                               shell=True,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)

                # If dcm2niix didn't work use dcm2nii
                # print path.join(dest_dir, bids_name + '.nii.gz')
                if not os.path.exists(path.join(bids_dest_dir,
                                                bids_name + '.nii.gz')) or not os.path.exists(
                        path.join(bids_dest_dir,
                                  bids_name + '.bvec') or not os.path.exists(
                            path.join(bids_dest_dir,
                                      bids_name + '.bval'))):
                    cprint('\tConversion with dcm2niix failed, trying with dcm2nii')

                    # Find all the files eventually created by dcm2niix and remove them
                    dwi_dcm2niix = glob(path.join(bids_dest_dir, bids_name + '*'))

                    for d in dwi_dcm2niix:
                        # print 'Removing the old', d
                        os.remove(d)

                    command = 'dcm2nii -a n -d n -e n -i y -g y -p n -m n -r n -x n -o ' + bids_dest_dir + ' ' + dwi_path
                    subprocess.run(command,
                                   shell=True,
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.DEVNULL)
                    nii_file = path.join(bids_dest_dir,
                                         subj.replace('_', '') + '.nii.gz')
                    bvec_file = path.join(bids_dest_dir,
                                          subj.replace('_', '') + '.bvec')
                    bval_file = path.join(bids_dest_dir,
                                          subj.replace('_', '') + '.bval')

                    if os.path.exists(bvec_file) and os.path.exists(
                            bval_file):
                        os.rename(bvec_file, path.join(bids_dest_dir,
                                                       bids_name + '.bvec'))
                        os.rename(bval_file, path.join(bids_dest_dir,
                                                       bids_name + '.bval'))
                    else:
                        cprint('WARNING: bvec and bval not generated by dcm2nii'
                               + ' for subject ' + subj + ' and session ' + ses)

                    if os.path.exists(nii_file):
                        os.rename(nii_file, path.join(bids_dest_dir,
                                                      bids_name + '.nii.gz'))
                    else:
                        cprint('WARNING: CONVERSION FAILED...'
                               + ' for subject ' + subj + ' and session ' + ses)


def dwi_image(subject_id, timepoint, visit_str, ida_meta_scans, mri_qc_subj, enhanced):
    """

    Args:
        subject_id:
        timepoint:
        visit_str:
        ida_meta_scans:
        mri_qc_subj:
        enhanced:

    Returns:

    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars

    sel_image = select_image_qc(list(ida_meta_scans.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = ida_meta_scans[ida_meta_scans.IMAGEUID == sel_image].iloc[0]

    sequence = sel_scan.Sequence
    sequence = replace_sequence_chars(sequence)

    image_dict = {'Subject_ID': subject_id,
                  'VISCODE': timepoint,
                  'Visit': visit_str,
                  'Sequence': sequence,
                  'Scan_Date': sel_scan['Scan Date'],
                  'Study_ID': str(int(sel_scan.LONISID)),
                  'Series_ID': str(int(sel_scan.LONIUID)),
                  'Image_ID': str(int(sel_scan.IMAGEUID)),
                  'Field_Strength': sel_scan.MagStrength,
                  'Scanner': sel_scan.Scanner,
                  'Enhanced': enhanced}

    return image_dict


def select_image_qc(id_list, mri_qc_subj):
    """

    Args:
        id_list:
        mri_qc_subj:

    Returns:

    """
    import numpy as np

    if len(id_list) == 0:
        return None

    selected_image = None
    image_ids = ['I' + str(imageuid) for imageuid in id_list]
    int_ids = [int(imageuid) for imageuid in id_list]
    images_qc = mri_qc_subj[mri_qc_subj.loni_image.isin(image_ids)]

    if images_qc.shape[0] < 1:
        return min(int_ids)

    if np.sum(images_qc.series_selected) == 1:
        selected_image = images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
    else:
        images_not_rejected = images_qc[images_qc.series_quality < 4]

        if images_not_rejected.shape[0] < 1:

            # There are no images that passed the qc
            # so we'll try to see if there are other images without qc,
            # otherwise return None
            qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
            no_qc_ids = list(set(int_ids) - qc_ids)

            if len(no_qc_ids) == 0:
                return None
            else:
                return min(no_qc_ids)

        series_quality = [q if q > 0 else 4 for q in list(images_not_rejected.series_quality)]
        best_q = np.amin(series_quality)
        if best_q == 4:
            best_q = -1
        images_best_qc = images_not_rejected[images_not_rejected.series_quality == best_q]
        if images_best_qc.shape[0] == 1:
            selected_image = images_best_qc.iloc[0].loni_image[1:]
        else:
            best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
            selected_image = min(best_ids)

    return int(selected_image)


def visits_to_timepoints_dwi(subject, ida_meta_subj, adnimerge_subj):
    """

    Args:
        subject:
        ida_meta_subj:
        adnimerge_subj:

    Returns:

    """
    from datetime import datetime
    from clinica.iotools.converters.adni_to_bids.adni_utils import days_between
    from clinica.utils.stream import cprint

    visits = dict()
    unique_visits = list(ida_meta_subj.Visit.unique())
    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():
        visit = adni_row[1]
        if visit.ORIGPROT == 'ADNI2':
            if visit.VISCODE == 'bl':
                preferred_visit_name = 'ADNI2 Screening MRI-New Pt'
            elif visit.VISCODE == 'm03':
                preferred_visit_name = 'ADNI2 Month 3 MRI-New Pt'
            elif visit.VISCODE == 'm06':
                preferred_visit_name = 'ADNI2 Month 6-New Pt'
            else:
                year = str(int(visit.VISCODE[1:]) / 12)
                preferred_visit_name = 'ADNI2 Year ' + year + ' Visit'
        else:
            if visit.VISCODE == 'bl':
                if visit.ORIGPROT == 'ADNI1':
                    preferred_visit_name = 'ADNI Screening'
                else:  # ADNIGO
                    preferred_visit_name = 'ADNIGO Screening MRI'
            elif visit.VISCODE == 'm03':  # Only for ADNIGO Month 3
                preferred_visit_name = 'ADNIGO Month 3 MRI'
            else:
                month = int(visit.VISCODE[1:])
                if month < 54:
                    preferred_visit_name = 'ADNI1/GO Month ' + str(month)
                else:
                    preferred_visit_name = 'ADNIGO Month ' + str(month)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits.keys():
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint('Multiple visits for one timepoint!')
                cprint(subject)
                cprint(key_preferred_visit)
                cprint(visits[key_preferred_visit])
                cprint(visit)
            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:
        image = (ida_meta_subj[ida_meta_subj.Visit == visit]).iloc[0]
        min_db = 100000
        min_db2 = 0
        min_visit = None
        min_visit2 = None

        for timepoint in pending_timepoints:
            db = days_between(image['Scan Date'], timepoint.EXAMDATE)
            if db < min_db:
                min_db2 = min_db
                min_visit2 = min_visit

                min_db = db
                min_visit = timepoint

        if min_visit is None:
            cprint('No corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.Visit)
            cprint(image)
            continue

        if min_visit2 is not None and min_db > 90:
            cprint('More than 60 days for corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.Visit + ' on ' + image.ScanDate)
            cprint('Timepoint 1: ' + min_visit.VISCODE + ' - ' + min_visit.ORIGPROT + ' on ' + min_visit.EXAMDATE + ' (Distance: ' + str(
                min_db) + ' days)')
            cprint('Timepoint 2: ' + min_visit2.VISCODE + ' - ' + min_visit2.ORIGPROT + ' on ' + min_visit2.EXAMDATE + ' (Distance: ' + str(
                min_db2) + ' days)')

            # If image is too close to the date between two visits we prefer the earlier visit
            if (datetime.strptime(min_visit.EXAMDATE, "%Y-%m-%d")
                    > datetime.strptime(image.ScanDate, "%Y-%m-%d")
                    > datetime.strptime(min_visit2.EXAMDATE, "%Y-%m-%d")):
                dif = days_between(min_visit.EXAMDATE, min_visit2.EXAMDATE)
                if abs((dif / 2.0) - min_db) < 30:
                    min_visit = min_visit2

            cprint('We prefer ' + min_visit.VISCODE)

        key_min_visit = (min_visit.VISCODE, min_visit.COLPROT, min_visit.ORIGPROT)
        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image.Visit
        elif visits[key_min_visit] != image.Visit:
            cprint('Multiple visits for one timepoint!')
            cprint(subject)
            cprint(key_min_visit)
            cprint(visits[key_min_visit])
            cprint(image.Visit)

    return visits


def visits_to_timepoints_dwi_refactoring(subject, ida_meta_subj, adnimerge_subj):
    """

    Args:
        subject:
        ida_meta_subj:
        adnimerge_subj:

    Returns:

    """
    from datetime import datetime
    from clinica.iotools.converters.adni_to_bids.adni_utils import days_between
    from clinica.utils.stream import cprint

    visits = dict()
    unique_visits = list(ida_meta_subj.VISIT.unique())
    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():
        visit = adni_row[1]
        if visit.ORIGPROT == 'ADNI2':
            if visit.VISCODE == 'bl':
                preferred_visit_name = 'ADNI2 Screening MRI-New Pt'
            elif visit.VISCODE == 'm03':
                preferred_visit_name = 'ADNI2 Month 3 MRI-New Pt'
            elif visit.VISCODE == 'm06':
                preferred_visit_name = 'ADNI2 Month 6-New Pt'
            else:
                year = str(int(visit.VISCODE[1:]) / 12)
                preferred_visit_name = 'ADNI2 Year ' + year + ' Visit'
        else:
            if visit.VISCODE == 'bl':
                if visit.ORIGPROT == 'ADNI1':
                    preferred_visit_name = 'ADNI Screening'
                else:  # ADNIGO
                    preferred_visit_name = 'ADNIGO Screening MRI'
            elif visit.VISCODE == 'm03':  # Only for ADNIGO Month 3
                preferred_visit_name = 'ADNIGO Month 3 MRI'
            else:
                month = int(visit.VISCODE[1:])
                if month < 54:
                    preferred_visit_name = 'ADNI1/GO Month ' + str(month)
                else:
                    preferred_visit_name = 'ADNIGO Month ' + str(month)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits.keys():
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint('Multiple visits for one timepoint!')
                cprint(subject)
                cprint(key_preferred_visit)
                cprint(visits[key_preferred_visit])
                cprint(visit)
            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:
        image = (ida_meta_subj[ida_meta_subj.VISIT == visit]).iloc[0]
        min_db = 100000
        min_db2 = 0
        min_visit = None
        min_visit2 = None

        for timepoint in pending_timepoints:
            db = days_between(image['SCANDATE'], timepoint.EXAMDATE)
            if db < min_db:
                min_db2 = min_db
                min_visit2 = min_visit

                min_db = db
                min_visit = timepoint

        if min_visit is None:
            cprint('No corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.VISIT)
            cprint(image)
            continue

        if min_visit2 is not None and min_db > 90:
            cprint('More than 60 days for corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.VISIT + ' on ' + image.SCANDATE)
            cprint('Timepoint 1: ' + min_visit.VISCODE + ' - ' + min_visit.ORIGPROT + ' on ' + min_visit.EXAMDATE + ' (Distance: ' + str(
                min_db) + ' days)')
            cprint('Timepoint 2: ' + min_visit2.VISCODE + ' - ' + min_visit2.ORIGPROT + ' on ' + min_visit2.EXAMDATE + ' (Distance: ' + str(
                min_db2) + ' days)')

            # If image is too close to the date between two visits we prefer the earlier visit
            if (datetime.strptime(min_visit.EXAMDATE, "%Y-%m-%d")
                    > datetime.strptime(image.SCANDATE, "%Y-%m-%d")
                    > datetime.strptime(min_visit2.EXAMDATE, "%Y-%m-%d")):
                dif = days_between(min_visit.EXAMDATE, min_visit2.EXAMDATE)
                if abs((dif / 2.0) - min_db) < 30:
                    min_visit = min_visit2

            cprint('We prefer ' + min_visit.VISCODE)

        key_min_visit = (min_visit.VISCODE, min_visit.COLPROT, min_visit.ORIGPROT)
        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image.VISIT
        elif visits[key_min_visit] != image.VISIT:
            cprint('Multiple visits for one timepoint!')
            # cprint(subject)
            # cprint(key_min_visit)
            # cprint(visits[key_min_visit])
            # cprint(image.Visit)

    return visits


def dwi_image_refactoring(subject_id, timepoint, visit_str, ida_meta_scans, mri_qc_subj, enhanced):
    """

    Args:
        subject_id:
        timepoint:
        visit_str:
        ida_meta_scans:
        mri_qc_subj:
        enhanced:

    Returns:

    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars

    sel_image = select_image_qc(list(ida_meta_scans.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = ida_meta_scans[ida_meta_scans.IMAGEUID == sel_image].iloc[0]

    sequence = sel_scan.SEQUENCE
    sequence = replace_sequence_chars(sequence)

    image_dict = {'Subject_ID': subject_id,
                  'VISCODE': timepoint,
                  'Visit': visit_str,
                  'Sequence': sequence,
                  'Scan_Date': sel_scan['SCANDATE'],
                  'Study_ID': str(int(sel_scan.STUDYID)),
                  'Series_ID': str(int(sel_scan.SERIESID)),
                  'Image_ID': str(int(sel_scan.IMAGEUID)),
                  'Field_Strength': sel_scan.MAGSTRENGTH,
                  'Enhanced': enhanced}

    return image_dict
